#!/usr/bin/env python
'''
Pandas DataFrame GUI

A minimalistic GUI for analyzing Pandas DataFrames based on wxPython.


Usage
-----
import datetime as dt
import numpy as np
import pandas as pd
filename = 'Hesse_DB2_1997-2011.csv'
# filename = '1.csv'
# not needed for US format
parser = lambda date: dt.datetime.strptime(date, '%Y-%m-%d %H:%M')
names = ['Date', 'FCO2 H1 (umol/m2.s)', 'FCO2_cor H1 (umol/m2.s)',
         'FH2O H1 (mmol/m2.s)', 'H H1 (W/m2)',
         'Rg_Kipp H1 (W/m2)', 'alb H1 (-)', 'Rnet_CNR1 H1 (W/m2)',
         'NDVI H1 (-)', 'Patm (hPa)',
         'T_Vais H1 (degC)', 'WS_EC H1 (m/s)', 'Prec H1 (mm/30min)']

# pandas
df = pd.read_csv(filename, ';', parse_dates=[0], date_parser=parser,
                 index_col=0, header=0, usecols=names)

# jams
from jams import fread, fsread, ascii2eng
names = names[1:]
head = fread(filename, skip=1, cname=names, header=True)
dat, date = fsread(filename, skip=1, cname=names, sname=['Date'],
                   squeeze=True, strarr=True, strip=False)
df = pd.DataFrame(dat, pd.to_datetime(ascii2eng(date)), head)

# This must be before any other call to matplotlib
# because it use the wxAgg backend.
# This means, do not use --pylab with ipython.
df.replace(-9999., np.nan, inplace=True)
from jams import dfgui
dfgui.show(df)


Features
--------
- Tabular view of data frame
- Columns are sortable (by clicking column header)
- Columns can be enabled/disabled (left click on 'Columns' tab)
- Columns can be rearranged (right click drag on 'Columns' tab)
- Generic filtering: Write arbitrary Python expression to filter rows.
  *Warning:* Uses Python's `eval` -- use with care.
  Select column. Use filter like:
      _ > 0.
      _ > datetime.datetime(1998,1,1)
- Histogram plots
- Scatter plots


Dependencies
------------
Needs wxpython.


License
-------
This file is part of the JAMS Python package,
distributed under the MIT License.

Copyright (c) 2019 Matthias Cuntz - mc (at) macu (dot) de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


History
-------
Written,  Fabian Keller (fabian.keller@blue-yonder.com), 2014,
              http://github.com/bluenote10/PandasDataFrameGUI
Modified, Matthias Cuntz, Jan 2019 - exclude NaN in histograms
                                   - add index as 'Date' column automatically
                                   - plot controls
          Matthias Cuntz, Feb 2019 - added 2nd line/scatter in scatter plots
                                     -> Bug: time axis does not work with first
                                             line but with second line.
                                   - label x-axis on histograms
          Matthias Cuntz, Jun 2019 - make application
                                   - Removed bug that x-axis in scatter plots
                                     is only correct if line2 is chosen.
          Matthias Cuntz, Jul 2019 - added example to filter panel
          Matthias Cuntz, Aug 2019 - added checkbox for linking the two y-axes
                                     in scatter plots
          Matthias Cuntz, Sep 2020 - added checkbox to invert second axis
          Matthias Cuntz, Sep 2020 - changed wx.ListItemAttr() to wx.ItemAttr()
          Matthias Cuntz, Apr 2021 - undef on command line
                                   - flake8 compatible
          Matthias Cuntz, May 2021 - sort columns with command line option -c
          Matthias Cuntz, May 2021 - write coordinates and values on bottom of
                                     plot
          Matthias Cuntz, Sep 2021 - catch if no infile given
          Matthias Cuntz, Jul 2022 - add header argument
          Matthias Cuntz, Jun 2023 - def main for entry_points in setup.cfg

'''
# --------------------------------------------------------------------
# import
#
# pip install wxpython
import numpy as np
import pandas as pd
# unused import required to allow 'eval' of date filters
import datetime as dt
from bisect import bisect
import wx
import matplotlib
matplotlib.use('WXAgg')
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.backends.backend_wx import NavigationToolbar2Wx
from matplotlib.figure import Figure
import matplotlib.dates as mpld
from matplotlib import pyplot as plt
plt.style.use('seaborn-v0_8-dark')


def format_coord_scatter(x, y, ax, ax2, xdtype, ydtype, y2dtype):
    """
    Formatter function for scatter plot with left and right axis
    having the same x-axis.

    Parameters
    ----------
    x, y : float
        Data coordinates of `ax2`.
    ax, ax2: matplotlib.axes._subplots.AxesSubplot
        Matplotlib axes object for left-hand and right-hand y-axis, resp.
    xdtype, ydtype, y2dtype: numpy.dtype
        Numpy dtype of data of x-values (xdtype), left-hand side y-values
        (ydtype), and right-hand side y-values (y2dtype)

    Returns
    -------
    String with left-hand side and right hand-side coordinates.

    Examples
    --------
    ax = plt.subplot(111)
    ax2 = ax.twinx()
    ax.plot(xx, yy)
    ax2.plot(xx, yy2)
    ax2.format_coord = lambda x, y: format_coord_scatter(
        x, y, ax, ax2, xx.dtype, yy.dtype, yy2.dtype)
    """
    # convert to display coords
    # https://stackoverflow.com/questions/21583965/matplotlib-cursor-value-with-two-axes
    display_coord = ax2.transData.transform((x, y))
    # convert back to data coords with respect to ax
    inv      = ax.transData.inverted()
    ax_coord = inv.transform(display_coord)

    # Special treatment for datetime
    # https://stackoverflow.com/questions/49267011/matplotlib-datetime-from-event-coordinates
    if xdtype == 'datetime64[ns]':
        xstr = mpld.num2date(x).strftime('%Y-%m-%d %H:%M')
    else:
        xstr  = '{:.3g}'.format(x)
    if ydtype == 'datetime64[ns]':
        ystr = mpld.num2date(ax_coord[1]).strftime('%Y-%m-%d %H:%M')
    else:
        ystr  = '{:.3g}'.format(ax_coord[1])
    if y2dtype == 'datetime64[ns]':
        y2str = mpld.num2date(y).strftime('%Y-%m-%d %H:%M')
    else:
        y2str = '{:.3g}'.format(y)
    out  = 'L: (' + xstr + ', ' + ystr + ')'
    out += ' R: (' + xstr + ', ' + y2str + ')'
    return out


# --------------------------------------------------------------------
# Basic List Class
#
class ListCtrlDataFrame(wx.ListCtrl):

    # TODO: we could do something more sophisticated to come
    #       up with a reasonable column width...
    DEFAULT_COLUMN_WIDTH = 100
    TMP_SELECTION_COLUMN = 'tmp_selection_column'

    def __init__(self, parent, df, status_bar_callback):
        wx.ListCtrl.__init__(self, parent, -1, style=wx.LC_REPORT |
                             wx.LC_VIRTUAL | wx.LC_HRULES | wx.LC_VRULES |
                             wx.LB_MULTIPLE)
        self.status_bar_callback = status_bar_callback

        self.df_orig = df
        self.original_columns = self.df_orig.columns[:]
        self.current_columns = self.df_orig.columns[:]

        self.sort_by_column = None

        self._reset_mask()

        # prepare attribute for alternating colors of rows
        # self.attr_light_blue = wx.ListItemAttr()
        self.attr_light_blue = wx.ItemAttr()
        self.attr_light_blue.SetBackgroundColour("#D6EBFF")

        self.Bind(wx.EVT_LIST_COL_CLICK, self._on_col_click)
        self.Bind(wx.EVT_RIGHT_DOWN, self._on_right_click)

        self.df = pd.DataFrame({})  # init empty to force initial update
        self._update_rows()
        self._update_columns(self.original_columns)

    def _reset_mask(self):
        # self.mask = [True] * self.df_orig.shape[0]
        self.mask = pd.Series([True] * self.df_orig.shape[0],
                              index=self.df_orig.index)

    def _update_columns(self, columns):
        self.ClearAll()
        for i, col in enumerate(columns):
            self.InsertColumn(i, col)
            self.SetColumnWidth(i, self.DEFAULT_COLUMN_WIDTH)
        # Note that we have to reset the count as well because ClearAll()
        # not only deletes columns but also the count...
        self.SetItemCount(len(self.df))

    def set_columns(self, columns_to_use):
        """
        External interface to set the column projections.
        """
        self.current_columns = columns_to_use
        self._update_rows()
        self._update_columns(columns_to_use)

    def _update_rows(self):
        old_len = len(self.df)
        self.df = self.df_orig.loc[self.mask.values, self.current_columns]
        new_len = len(self.df)
        if old_len != new_len:
            self.SetItemCount(new_len)
            self.status_bar_callback(0, "Number of rows: {}".format(new_len))

    def apply_filter(self, conditions):
        """
        External interface to set a filter.
        """
        old_mask = self.mask.copy()

        if len(conditions) == 0:
            self._reset_mask()

        else:
            self._reset_mask()  # set all to True for destructive conjunction

            no_error = True
            for column, condition in conditions:
                if condition.strip() == '':
                    continue
                condition = condition.replace(
                    "_", "self.df_orig['{}']".format(column))
                print("Evaluating condition:", condition)
                try:
                    tmp_mask = eval(condition)
                    if ( isinstance(tmp_mask, pd.Series) and
                         (tmp_mask.dtype == np.bool)):
                        self.mask &= tmp_mask
                except Exception as e:
                    print("Failed with:", e)
                    no_error = False
                    self.status_bar_callback(
                        1,
                        "Evaluating '{}' failed with: {}".format(condition, e)
                    )

            if no_error:
                self.status_bar_callback(1, "")

        has_changed = any(old_mask != self.mask)
        if has_changed:
            self._update_rows()

        return len(self.df), has_changed

    def get_selected_items(self):
        """
        Gets the selected items for the list control.
        Selection is returned as a list of selected indices,
        low to high.
        """
        selection = []
        current = -1    # start at -1 to get the first selected item
        while True:
            next = self.GetNextItem(current, wx.LIST_NEXT_ALL,
                                    wx.LIST_STATE_SELECTED)
            if next == -1:
                return selection
            else:
                selection.append(next)
                current = next

    def get_filtered_df(self):
        return self.df_orig.loc[self.mask, :]

    def _on_col_click(self, event):
        """
        Sort data frame by selected column.
        """
        # get currently selected items
        selected = self.get_selected_items()

        # append a temporary column to store the currently selected items
        self.df[self.TMP_SELECTION_COLUMN] = False
        self.df.iloc[selected, -1] = True

        # get column name to use for sorting
        col = event.GetColumn()

        # determine if ascending or descending
        if self.sort_by_column is None or self.sort_by_column[0] != col:
            ascending = True
        else:
            ascending = not self.sort_by_column[1]

        # store sort column and sort direction
        self.sort_by_column = (col, ascending)

        try:
            # pandas 0.17
            self.df.sort_values(self.df.columns[col], inplace=True,
                                ascending=ascending)
        except AttributeError:
            # pandas 0.16 compatibility
            self.df.sort(self.df.columns[col], inplace=True,
                         ascending=ascending)

        # deselect all previously selected
        for i in selected:
            self.Select(i, on=False)

        # determine indices of selection after sorting
        selected_bool = self.df.iloc[:, -1]  # == True
        selected = self.df.reset_index().index[selected_bool]

        # select corresponding rows
        for i in selected:
            self.Select(i, on=True)

        # delete temporary column
        del self.df[self.TMP_SELECTION_COLUMN]

    def _on_right_click(self, event):
        """
        Copies a cell into clipboard on right click. Unfortunately,
        determining the clicked column is not straightforward. This
        appraoch is inspired by the TextEditMixin in:
        /usr/lib/python2.7/dist-packages/wx-2.8-gtk2-unicode/wx/lib/mixins/listctrl.py
        More references:
        http://wxpython-users.1045709.n5.nabble.com/Getting-row-col-of-selected-cell-in-ListCtrl-td2360831.html
        https://groups.google.com/forum/#!topic/wxpython-users/7BNl9TA5Y5U
        https://groups.google.com/forum/#!topic/wxpython-users/wyayJIARG8c
        """
        if self.HitTest(event.GetPosition()) != wx.NOT_FOUND:
            x, y = event.GetPosition()
            row, flags = self.HitTest((x, y))

            col_locs = [0]
            loc = 0
            for n in range(self.GetColumnCount()):
                loc = loc + self.GetColumnWidth(n)
                col_locs.append(loc)

            scroll_pos = self.GetScrollPos(wx.HORIZONTAL)
            # this is crucial step to get the scroll pixel units
            unit_x, unit_y = self.GetMainWindow().GetScrollPixelsPerUnit()

            col = bisect(col_locs, x + scroll_pos * unit_x) - 1

            value = self.df.iloc[row, col]
            # print(row, col, scroll_pos, value)

            clipdata = wx.TextDataObject()
            clipdata.SetText(str(value))
            wx.TheClipboard.Open()
            wx.TheClipboard.SetData(clipdata)
            wx.TheClipboard.Close()

    def OnGetItemText(self, item, col):
        """
        Implements the item getter for a "virtual" ListCtrl.
        """
        value = self.df.iloc[item, col]
        # print("retrieving %d %d %s" % (item, col, value))
        return str(value)

    def OnGetItemAttr(self, item):
        """
        Implements the attribute getter for a "virtual" ListCtrl.
        """
        if item % 2 == 0:
            return self.attr_light_blue
        else:
            return None


# --------------------------------------------------------------------
# Date table panel
#
class DataframePanel(wx.Panel):
    """
    Panel providing the main data frame table view.
    """
    def __init__(self, parent, df, status_bar_callback):
        wx.Panel.__init__(self, parent)

        self.df_list_ctrl = ListCtrlDataFrame(self, df, status_bar_callback)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.df_list_ctrl, 1, wx.ALL | wx.EXPAND | wx.GROW, 5)
        self.SetSizer(sizer)
        self.Show()


class ListBoxDraggable(wx.ListBox):
    """
    Helper class to provide ListBox with extended behavior.
    """
    def __init__(self, parent, size, data, *args, **kwargs):

        wx.ListBox.__init__(self, parent, size, **kwargs)

        self.data = data

        self.InsertItems(data, 0)

        self.Bind(wx.EVT_LISTBOX, self.on_selection_changed)

        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)

        self.Bind(wx.EVT_RIGHT_DOWN, self.on_right_down)
        self.Bind(wx.EVT_RIGHT_UP, self.on_right_up)
        self.Bind(wx.EVT_MOTION, self.on_move)

        self.index_iter = range(len(self.data))

        self.selected_items = [True] * len(self.data)
        self.index_mapping = list(range(len(self.data)))

        self.drag_start_index = None

        self.update_selection()
        self.SetFocus()

    def on_left_down(self, event):
        if self.HitTest(event.GetPosition()) != wx.NOT_FOUND:
            index = self.HitTest(event.GetPosition())
            self.selected_items[index] = not self.selected_items[index]
            # doesn't really work to update selection direclty (focus issues)
            # instead we wait for the EVT_LISTBOX event and fix the selection
            # there...
            # self.update_selection()
            # TODO: we could probably use wx.CallAfter
        event.Skip()

    def update_selection(self):
        # self.SetFocus()
        # print(self.selected_items)
        for i in self.index_iter:
            if self.IsSelected(i) and not self.selected_items[i]:
                # print("Deselecting", i)
                self.Deselect(i)
            elif not self.IsSelected(i) and self.selected_items[i]:
                # print("Selecting", i)
                self.Select(i)

    def on_selection_changed(self, evt):
        self.update_selection()
        evt.Skip()

    def on_right_down(self, event):
        if self.HitTest(event.GetPosition()) != wx.NOT_FOUND:
            index = self.HitTest(event.GetPosition())
            self.drag_start_index = index

    def on_right_up(self, event):
        self.drag_start_index = None
        event.Skip()

    def on_move(self, event):
        if self.drag_start_index is not None:
            if self.HitTest(event.GetPosition()) != wx.NOT_FOUND:
                index = self.HitTest(event.GetPosition())
                if self.drag_start_index != index:
                    self.swap(self.drag_start_index, index)
                    self.drag_start_index = index

    def swap(self, i, j):
        self.index_mapping[i], self.index_mapping[j] = (
            self.index_mapping[j], self.index_mapping[i])
        self.SetString(i, self.data[self.index_mapping[i]])
        self.SetString(j, self.data[self.index_mapping[j]])
        self.selected_items[i], self.selected_items[j] = (
            self.selected_items[j], self.selected_items[i])
        # self.update_selection()
        # print("Updated mapping:", self.index_mapping)
        new_event = wx.PyCommandEvent(wx.EVT_LISTBOX.typeId, self.GetId())
        self.GetEventHandler().ProcessEvent(new_event)

    def get_selected_data(self):
        selected = []
        for i, col in enumerate(self.data):
            if self.IsSelected(i):
                index = self.index_mapping[i]
                value = self.data[index]
                selected.append(value)
        # print("Selected data:", selected)
        return selected


# --------------------------------------------------------------------
# Column selection panel
#
class ColumnSelectionPanel(wx.Panel):
    """
    Panel for selecting and re-arranging columns.
    """
    def __init__(self, parent, columns, df_list_ctrl):
        wx.Panel.__init__(self, parent)

        self.columns = columns
        self.df_list_ctrl = df_list_ctrl

        self.list_box = ListBoxDraggable(self, -1, columns,
                                         style=wx.LB_EXTENDED)
        self.Bind(wx.EVT_LISTBOX, self.update_selected_columns)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.list_box, 1, wx.ALL | wx.EXPAND | wx.GROW, 5)
        self.SetSizer(sizer)
        self.list_box.SetFocus()

    def update_selected_columns(self, evt):
        selected = self.list_box.get_selected_data()
        self.df_list_ctrl.set_columns(selected)


# --------------------------------------------------------------------
# Filter panel
#
class FilterPanel(wx.Panel):
    """
    Panel for defining filter expressions.
    """
    def __init__(self, parent, columns, df_list_ctrl, change_callback):
        wx.Panel.__init__(self, parent)

        columns_with_neutral_selection = [''] + list(columns)
        self.columns = columns
        self.df_list_ctrl = df_list_ctrl
        self.change_callback = change_callback

        self.num_filters = 10

        self.main_sizer = wx.BoxSizer(wx.VERTICAL)

        self.example = wx.StaticText(
            self, wx.ID_ANY,
            label=("Select column. Use filter like:   _ > 0.   or   _ >"
                   " datetime.datetime(1998,1,1)"),
            size=wx.Size(560, 20))
        row_sizer = wx.BoxSizer(wx.HORIZONTAL)
        # row_sizer.Add(self.example, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        row_sizer.Add(self.example, 0, wx.ALL, 1)
        self.main_sizer.Add(row_sizer, 0, wx.EXPAND)

        self.combo_boxes = []
        self.text_controls = []

        for i in range(self.num_filters):
            combo_box = wx.ComboBox(self,
                                    choices=columns_with_neutral_selection,
                                    style=wx.CB_READONLY)
            text_ctrl = wx.TextCtrl(self, wx.ID_ANY, '')

            self.Bind(wx.EVT_COMBOBOX, self.on_combo_box_select)
            self.Bind(wx.EVT_TEXT, self.on_text_change)

            row_sizer = wx.BoxSizer(wx.HORIZONTAL)
            row_sizer.Add(combo_box, 0, wx.ALL, 5)
            # row_sizer.Add(text_ctrl, 1, wx.ALL | wx.EXPAND |
            #               wx.ALIGN_RIGHT, 5)
            row_sizer.Add(text_ctrl, 1, wx.ALL | wx.EXPAND, 5)

            self.combo_boxes.append(combo_box)
            self.text_controls.append(text_ctrl)
            self.main_sizer.Add(row_sizer, 0, wx.EXPAND)

        self.SetSizer(self.main_sizer)

    def on_combo_box_select(self, event):
        self.update_conditions()

    def on_text_change(self, event):
        self.update_conditions()

    def update_conditions(self):
        # print("Updating conditions")
        conditions = []
        for i in range(self.num_filters):
            column_index = self.combo_boxes[i].GetSelection()
            condition = self.text_controls[i].GetValue()
            if column_index != wx.NOT_FOUND and column_index != 0:
                # since we have added a dummy column for "deselect",
                # we have to subtract one
                column = self.columns[column_index - 1]
                conditions += [(column, condition)]
        num_matching, has_changed = self.df_list_ctrl.apply_filter(conditions)
        if has_changed:
            self.change_callback()
        # print("Num matching:", num_matching)


# --------------------------------------------------------------------
# Histogram plot panel
#
class HistogramPlot(wx.Panel):
    """
    Panel providing a histogram plot.
    """
    def __init__(self, parent, columns, df_list_ctrl):
        wx.Panel.__init__(self, parent)

        columns_with_neutral_selection = [''] + list(columns)
        self.columns = columns
        self.df_list_ctrl = df_list_ctrl

        self.figure = Figure(facecolor="white", figsize=(1, 1))
        self.axes = self.figure.add_subplot(111)
        self.canvas = FigureCanvas(self, -1, self.figure)

        chart_toolbar = NavigationToolbar2Wx(self.canvas)

        self.combo_box1 = wx.ComboBox(self,
                                      choices=columns_with_neutral_selection,
                                      style=wx.CB_READONLY)

        self.Bind(wx.EVT_COMBOBOX, self.on_combo_box_select)

        row_sizer = wx.BoxSizer(wx.HORIZONTAL)
        # row_sizer.Add(self.combo_box1, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        row_sizer.Add(self.combo_box1, 0, wx.ALL, 5)
        row_sizer.Add(chart_toolbar, 0, wx.ALL, 5)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.canvas, 1, flag=wx.EXPAND, border=5)
        sizer.Add(row_sizer)
        self.SetSizer(sizer)

    def on_combo_box_select(self, event):
        self.redraw()

    def redraw(self):
        column_index1 = self.combo_box1.GetSelection()
        if column_index1 != wx.NOT_FOUND and column_index1 != 0:
            # subtract one to remove the neutral selection index
            column_index1 -= 1

            df = self.df_list_ctrl.get_filtered_df()

            if len(df) > 0:
                self.axes.clear()

                column = df.iloc[:, column_index1]
                is_string_col = ((column.dtype == np.object) and
                                 isinstance(column.values[0], str))
                if is_string_col:
                    value_counts = column.value_counts().sort_index()
                    value_counts.plot(kind='bar', ax=self.axes)
                else:
                    self.axes.hist(
                        np.ma.array(column.values,
                                    mask=~np.isfinite(column.values)),
                        bins=100)
                    self.axes.xaxis.set_label_text(self.columns[column_index1])

                self.canvas.draw()


# --------------------------------------------------------------------
# Scatter plot panel
#
class ScatterPlot(wx.Panel):
    """
    Panel providing a scatter plot.
    """
    def __init__(self, parent, columns, df_list_ctrl):
        wx.Panel.__init__(self, parent)

        columns_with_neutral_selection = [''] + list(columns)
        self.columns = columns
        self.df_list_ctrl = df_list_ctrl

        self.figure = Figure(facecolor="white", figsize=(1, 1))
        self.axes  = self.figure.add_subplot(111)
        self.axes2 = self.axes.twinx()
        self.canvas = FigureCanvas(self, -1, self.figure)
        # self.canvas.mpl_connect('pick_event', self.onpick)

        chart_toolbar = NavigationToolbar2Wx(self.canvas)

        # first line
        self.combo_box1 = wx.ComboBox(self,
                                      choices=columns_with_neutral_selection,
                                      style=wx.CB_READONLY)
        self.combo_box2 = wx.ComboBox(self,
                                      choices=columns_with_neutral_selection,
                                      style=wx.CB_READONLY)

        self.linestyle_label = wx.StaticText(self, wx.ID_ANY,
                                             label="linestyle",
                                             size=wx.Size(60, 20))
        self.linestyle = wx.TextCtrl(self, wx.ID_ANY, value="-",
                                     size=wx.Size(40, 20))
        self.linewidth_label = wx.StaticText(self, wx.ID_ANY,
                                             label="linewidth",
                                             size=wx.Size(60, 20))
        self.linewidth = wx.TextCtrl(self, wx.ID_ANY, value="1",
                                     size=wx.Size(25, 20))
        self.linecolor_label = wx.StaticText(self, wx.ID_ANY,
                                             label="linecolor",
                                             size=wx.Size(60, 20))
        self.linecolor = wx.TextCtrl(self, wx.ID_ANY, value="b",
                                     size=wx.Size(40, 20))
        self.marker_label = wx.StaticText(self, wx.ID_ANY, label="marker",
                                          size=wx.Size(50, 20))
        self.marker = wx.TextCtrl(self, wx.ID_ANY, value="None",
                                  size=wx.Size(40, 20))
        self.markersize_label = wx.StaticText(self, wx.ID_ANY,
                                              label="markersize",
                                              size=wx.Size(70, 20))
        self.markersize = wx.TextCtrl(self, wx.ID_ANY, value="1",
                                      size=wx.Size(25, 20))
        self.markerfacecolor_label = wx.StaticText(self, wx.ID_ANY,
                                                   label="markerfacecolor",
                                                   size=wx.Size(100, 20))
        self.markerfacecolor = wx.TextCtrl(self, wx.ID_ANY, value="b",
                                           size=wx.Size(40, 20))
        self.markeredgecolor_label = wx.StaticText(self, wx.ID_ANY,
                                                   label="markeredgecolor",
                                                   size=wx.Size(110, 20))
        self.markeredgecolor = wx.TextCtrl(self, wx.ID_ANY, value="b",
                                           size=wx.Size(40, 20))
        self.markeredgewidth_label = wx.StaticText(self, wx.ID_ANY,
                                                   label="markeredgewidth",
                                                   size=wx.Size(110, 20))
        self.markeredgewidth = wx.TextCtrl(self, wx.ID_ANY, value="1",
                                           size=wx.Size(25, 20))

        row_sizer1 = wx.BoxSizer(wx.HORIZONTAL)
        # row_sizer1.Add(self.combo_box1, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer1.Add(self.combo_box2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        row_sizer1.Add(self.combo_box1, 0, wx.ALL, 5)
        row_sizer1.Add(self.combo_box2, 0, wx.ALL, 5)
        row_sizer1.Add(chart_toolbar, 0, wx.ALL, 5)

        row_sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        # row_sizer2.Add(self.linestyle_label, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer2.Add(self.linestyle, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer2.Add(self.linewidth_label, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer2.Add(self.linewidth, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer2.Add(self.linecolor_label, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer2.Add(self.linecolor, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        row_sizer2.Add(self.linestyle_label, 0, wx.ALL, 1)
        row_sizer2.Add(self.linestyle, 0, wx.ALL, 5)
        row_sizer2.Add(self.linewidth_label, 0, wx.ALL, 1)
        row_sizer2.Add(self.linewidth, 0, wx.ALL, 5)
        row_sizer2.Add(self.linecolor_label, 0, wx.ALL, 1)
        row_sizer2.Add(self.linecolor, 0, wx.ALL, 5)

        row_sizer3 = wx.BoxSizer(wx.HORIZONTAL)
        # row_sizer3.Add(self.marker_label, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer3.Add(self.marker, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer3.Add(self.markersize_label, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer3.Add(self.markersize, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer3.Add(self.markerfacecolor_label, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer3.Add(self.markerfacecolor, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer3.Add(self.markeredgecolor_label, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer3.Add(self.markeredgecolor, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer3.Add(self.markeredgewidth_label, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer3.Add(self.markeredgewidth, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 5)
        row_sizer3.Add(self.marker_label, 0, wx.ALL, 1)
        row_sizer3.Add(self.marker, 0, wx.ALL, 5)
        row_sizer3.Add(self.markersize_label, 0, wx.ALL, 1)
        row_sizer3.Add(self.markersize, 0, wx.ALL, 5)
        row_sizer3.Add(self.markerfacecolor_label, 0, wx.ALL, 1)
        row_sizer3.Add(self.markerfacecolor, 0, wx.ALL, 5)
        row_sizer3.Add(self.markeredgecolor_label, 0, wx.ALL, 1)
        row_sizer3.Add(self.markeredgecolor, 0, wx.ALL, 5)
        row_sizer3.Add(self.markeredgewidth_label, 0, wx.ALL, 1)
        row_sizer3.Add(self.markeredgewidth, 0, wx.ALL, 5)

        # second line
        self.combo_box3 = wx.ComboBox(self,
                                      choices=columns_with_neutral_selection,
                                      style=wx.CB_READONLY)
        self.checkbox1 = wx.CheckBox(self, label="Same y-axis")
        self.checkbox2 = wx.CheckBox(self, label="Invert y-axis")

        self.linestyle_label2 = wx.StaticText(self, wx.ID_ANY,
                                              label="linestyle",
                                              size=wx.Size(60, 20))
        self.linestyle2 = wx.TextCtrl(self, wx.ID_ANY, value="--",
                                      size=wx.Size(40, 20))
        self.linewidth_label2 = wx.StaticText(self, wx.ID_ANY,
                                              label="linewidth",
                                              size=wx.Size(60, 20))
        self.linewidth2 = wx.TextCtrl(self, wx.ID_ANY, value="1",
                                      size=wx.Size(25, 20))
        self.linecolor_label2 = wx.StaticText(self, wx.ID_ANY,
                                              label="linecolor",
                                              size=wx.Size(60, 20))
        self.linecolor2 = wx.TextCtrl(self, wx.ID_ANY, value="r",
                                      size=wx.Size(40, 20))
        self.marker_label2 = wx.StaticText(self, wx.ID_ANY, label="marker",
                                           size=wx.Size(50, 20))
        self.marker2 = wx.TextCtrl(self, wx.ID_ANY, value="None",
                                   size=wx.Size(40, 20))
        self.markersize_label2 = wx.StaticText(self, wx.ID_ANY,
                                               label="markersize",
                                               size=wx.Size(70, 20))
        self.markersize2 = wx.TextCtrl(self, wx.ID_ANY, value="1",
                                       size=wx.Size(25, 20))
        self.markerfacecolor_label2 = wx.StaticText(self, wx.ID_ANY,
                                                    label="markerfacecolor",
                                                    size=wx.Size(100, 20))
        self.markerfacecolor2 = wx.TextCtrl(self, wx.ID_ANY, value="r",
                                            size=wx.Size(40, 20))
        self.markeredgecolor_label2 = wx.StaticText(self, wx.ID_ANY,
                                                    label="markeredgecolor",
                                                    size=wx.Size(110, 20))
        self.markeredgecolor2 = wx.TextCtrl(self, wx.ID_ANY, value="r",
                                            size=wx.Size(40, 20))
        self.markeredgewidth_label2 = wx.StaticText(self, wx.ID_ANY,
                                                    label="markeredgewidth",
                                                    size=wx.Size(110, 20))
        self.markeredgewidth2 = wx.TextCtrl(self, wx.ID_ANY, value="1",
                                            size=wx.Size(25, 20))

        row_sizer4 = wx.BoxSizer(wx.HORIZONTAL)
        # row_sizer4.Add(self.combo_box3, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer4.Add(self.checkbox1, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        row_sizer4.Add(self.combo_box3, 0, wx.ALL, 5)
        row_sizer4.Add(self.checkbox1, 0, wx.ALL, 5)
        row_sizer4.Add(self.checkbox2, 0, wx.ALL, 5)

        row_sizer5 = wx.BoxSizer(wx.HORIZONTAL)
        # row_sizer5.Add(self.linestyle_label2, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer5.Add(self.linestyle2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer5.Add(self.linewidth_label2, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer5.Add(self.linewidth2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer5.Add(self.linecolor_label2, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer5.Add(self.linecolor2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        row_sizer5.Add(self.linestyle_label2, 0, wx.ALL, 1)
        row_sizer5.Add(self.linestyle2, 0, wx.ALL, 5)
        row_sizer5.Add(self.linewidth_label2, 0, wx.ALL, 1)
        row_sizer5.Add(self.linewidth2, 0, wx.ALL, 5)
        row_sizer5.Add(self.linecolor_label2, 0, wx.ALL, 1)
        row_sizer5.Add(self.linecolor2, 0, wx.ALL, 5)

        row_sizer6 = wx.BoxSizer(wx.HORIZONTAL)
        # row_sizer6.Add(self.marker_label2, 0, wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer6.Add(self.marker2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer6.Add(self.markersize_label2, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer6.Add(self.markersize2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer6.Add(self.markerfacecolor_label2, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer6.Add(self.markerfacecolor2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer6.Add(self.markeredgecolor_label2, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer6.Add(self.markeredgecolor2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        # row_sizer6.Add(self.markeredgewidth_label2, 0,
        #                wx.ALL | wx.ALIGN_CENTER, 1)
        # row_sizer6.Add(self.markeredgewidth2, 0, wx.ALL | wx.ALIGN_CENTER, 5)
        row_sizer6.Add(self.marker_label2, 0, wx.ALL, 1)
        row_sizer6.Add(self.marker2, 0, wx.ALL, 5)
        row_sizer6.Add(self.markersize_label2, 0, wx.ALL, 1)
        row_sizer6.Add(self.markersize2, 0, wx.ALL, 5)
        row_sizer6.Add(self.markerfacecolor_label2, 0, wx.ALL, 1)
        row_sizer6.Add(self.markerfacecolor2, 0, wx.ALL, 5)
        row_sizer6.Add(self.markeredgecolor_label2, 0, wx.ALL, 1)
        row_sizer6.Add(self.markeredgecolor2, 0, wx.ALL, 5)
        row_sizer6.Add(self.markeredgewidth_label2, 0, wx.ALL, 1)
        row_sizer6.Add(self.markeredgewidth2, 0, wx.ALL, 5)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.canvas, 1, flag=wx.EXPAND, border=5)
        sizer.Add(row_sizer1)
        sizer.Add(row_sizer2)
        sizer.Add(row_sizer3)
        sizer.Add(row_sizer4)
        sizer.Add(row_sizer5)
        sizer.Add(row_sizer6)
        self.SetSizer(sizer)

        self.Bind(wx.EVT_COMBOBOX, self.on_combo_box_select)
        self.Bind(wx.EVT_TEXT, self.on_text_change)
        self.Bind(wx.EVT_CHECKBOX, self.on_checkbox)

    def on_combo_box_select(self, event):
        self.redraw()

    def on_text_change(self, event):
        self.redraw()

    def on_checkbox(self, event):
        self.redraw()

    # def onpick(self, event):
    #     print('in pick')
    #     if self.canvas.HasCapture():
    #         self.canvas.ReleaseMouse()
    #     thisline = event.artist
    #     xdata = thisline.get_xdata()
    #     ydata = thisline.get_ydata()
    #     ind = event.ind
    #     points = tuple(zip(xdata[ind], ydata[ind]))
    #     print('onpick points:', points)

    def redraw(self, event=None):
        column_index1 = self.combo_box1.GetSelection()
        column_index2 = self.combo_box2.GetSelection()
        linestyle       = self.linestyle.GetLineText(0)
        linewidth       = float(self.linewidth.GetLineText(0))
        linecolor       = self.linecolor.GetLineText(0)
        marker          = self.marker.GetLineText(0)
        markersize      = float(self.markersize.GetLineText(0))
        markerfacecolor = self.markerfacecolor.GetLineText(0)
        markeredgecolor = self.markeredgecolor.GetLineText(0)
        markeredgewidth = float(self.markeredgewidth.GetLineText(0))
        # line 2
        column_index3 = self.combo_box3.GetSelection()
        linestyle2       = self.linestyle2.GetLineText(0)
        linewidth2       = float(self.linewidth2.GetLineText(0))
        linecolor2       = self.linecolor2.GetLineText(0)
        marker2          = self.marker2.GetLineText(0)
        markersize2      = float(self.markersize2.GetLineText(0))
        markerfacecolor2 = self.markerfacecolor2.GetLineText(0)
        markeredgecolor2 = self.markeredgecolor2.GetLineText(0)
        markeredgewidth2 = float(self.markeredgewidth2.GetLineText(0))
        sameyaxes   = self.checkbox1.GetValue()
        invertyaxis = self.checkbox2.GetValue()
        if ((column_index1 != wx.NOT_FOUND and column_index2 != wx.NOT_FOUND
             and column_index1 != 0 and column_index2 != 0) or
            (column_index1 != wx.NOT_FOUND and column_index3 != wx.NOT_FOUND
             and column_index1 != 0 and column_index3 != 0)):
            df = self.df_list_ctrl.get_filtered_df()
            if len(df) > 0:
                # subtract one to remove the neutral selection index
                column_index1 -= 1
                x  = df.iloc[:, column_index1].values
                column_index2 -= 1
                y  = df.iloc[:, column_index2].values
                column_index3 -= 1
                y2 = df.iloc[:, column_index3].values
                # It looks like using pandas dataframe.plot causes something
                # weird to crash in wx internally. Therefore we use plain
                # axes.plot functionality.
                # column_name1 = self.columns[column_index1]
                # column_name2 = self.columns[column_index2]
                # df.plot(kind='scatter', x=column_name1, y=column_name2)
                # Clear both axes first, otherwise x-axis only shows if line2
                # is chosen
                self.axes.clear()
                self.axes2.clear()
                ylim1 = [None, None]
                ylim2 = [None, None]
                if (column_index1 >= 0 and column_index2 >= 0):
                    self.axes.plot(x, y, linestyle=linestyle,
                                   linewidth=linewidth, color=linecolor,
                                   marker=marker, markersize=markersize,
                                   markerfacecolor=markerfacecolor,
                                   markeredgecolor=markeredgecolor,
                                   markeredgewidth=markeredgewidth)  # ,
                                   # picker=True, pickradius=15)
                    self.axes.spines['left'].set_color(linecolor)
                    self.axes.tick_params(axis='y', colors=linecolor)
                    self.axes.xaxis.set_label_text(self.columns[column_index1])
                    self.axes.yaxis.label.set_color(linecolor)
                    self.axes.yaxis.set_label_text(self.columns[column_index2])
                    ylim1 = self.axes.get_ylim()
                if (column_index1 >= 0 and column_index3 >= 0):
                    self.axes2.plot(x, y2, linestyle=linestyle2,
                                    linewidth=linewidth2, color=linecolor2,
                                    marker=marker2, markersize=markersize2,
                                    markerfacecolor=markerfacecolor2,
                                    markeredgecolor=markeredgecolor2,
                                    markeredgewidth=markeredgewidth2)  # ,
                                    # picker=True, pickradius=15)
                    self.axes2.spines['right'].set_color(linecolor2)
                    self.axes2.tick_params(axis='y', colors=linecolor2)
                    self.axes2.xaxis.set_label_text(
                        self.columns[column_index1])
                    self.axes2.yaxis.label.set_color(linecolor2)
                    self.axes2.yaxis.set_label_text(
                        self.columns[column_index3])
                    ylim2 = self.axes2.get_ylim()
                self.axes2.yaxis.set_label_position("right")
                self.axes2.yaxis.tick_right()
                self.axes2.format_coord = lambda xx, yy: format_coord_scatter(
                    xx, yy, self.axes, self.axes2, x.dtype, y.dtype, y2.dtype)
                if sameyaxes:
                    if (ylim1[0] is not None) and (ylim2[0] is not None):
                        ymin = min(ylim1[0], ylim2[0])
                    else:
                        if (ylim1[0] is not None):
                            ymin = ylim1[0]
                        else:
                            ymin = ylim2[0]
                    if (ylim1[1] is not None) and (ylim2[1] is not None):
                        ymax = max(ylim1[1], ylim2[1])
                    else:
                        if (ylim1[1] is not None):
                            ymax = ylim1[1]
                        else:
                            ymax = ylim2[1]
                    if (ymin is not None) and (ymax is not None):
                        self.axes.set_ylim([ymin, ymax])
                        self.axes2.set_ylim([ymin, ymax])
                elif invertyaxis:
                    if (ylim2[0] is not None):
                        ylim2 = ylim2[::-1]
                        self.axes2.set_ylim(ylim2)
                self.canvas.draw()


# --------------------------------------------------------------------
# Main window
#
class MainFrame(wx.Frame):
    """
    The main GUI window.
    """
    def __init__(self, df):
        wx.Frame.__init__(self, None, -1, "Pandas DataFrame GUI")

        # Here we create a panel and a notebook on the panel
        p = wx.Panel(self)
        nb = wx.Notebook(p)
        self.nb = nb

        # add index as Date column
        df.insert(0, 'Date', pd.Series(pd.to_datetime(df.index),
                                       index=df.index))

        columns = df.columns[:]

        self.CreateStatusBar(2, style=0)
        self.SetStatusWidths([200, -1])

        # create the page windows as children of the notebook
        self.page1 = DataframePanel(nb, df, self.status_bar_callback)
        self.page2 = ColumnSelectionPanel(nb, columns, self.page1.df_list_ctrl)
        self.page3 = FilterPanel(nb, columns, self.page1.df_list_ctrl,
                                 self.selection_change_callback)
        self.page4 = HistogramPlot(nb, columns, self.page1.df_list_ctrl)
        self.page5 = ScatterPlot(nb, columns, self.page1.df_list_ctrl)

        # add the pages to the notebook with the label to show on the tab
        nb.AddPage(self.page1, "Data Frame")
        nb.AddPage(self.page2, "Columns")
        nb.AddPage(self.page3, "Filters")
        nb.AddPage(self.page4, "Histogram")
        nb.AddPage(self.page5, "Scatter Plot")

        nb.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self.on_tab_change)

        # finally, put the notebook in a sizer for the panel to manage
        # the layout
        sizer = wx.BoxSizer()
        sizer.Add(nb, 1, wx.EXPAND)
        p.SetSizer(sizer)

        self.SetSize((1000, 800))
        self.Center()

    def on_tab_change(self, event):
        self.page2.list_box.SetFocus()
        page_to_select = event.GetSelection()
        wx.CallAfter(self.fix_focus, page_to_select)
        event.Skip(True)

    def fix_focus(self, page_to_select):
        page = self.nb.GetPage(page_to_select)
        page.SetFocus()
        if isinstance(page, DataframePanel):
            self.page1.df_list_ctrl.SetFocus()
        elif isinstance(page, ColumnSelectionPanel):
            self.page2.list_box.SetFocus()

    def status_bar_callback(self, i, new_text):
        self.SetStatusText(new_text, i)

    def selection_change_callback(self):
        self.page4.redraw()
        self.page5.redraw()


# --------------------------------------------------------------------
# Calling function
#
def show(df):
    """
    The main function to start the data frame GUI.
    """
    app = wx.App(False)
    frame = MainFrame(df)
    frame.Show()
    app.MainLoop()


# --------------------------------------------------------------------
# Script
#

def main(iargs=None):

    import argparse

    csort = False
    form  = '%Y-%m-%d %H:%M:%S'
    header = '0'
    sep   = None
    undef = None
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='''
A minimalistic GUI for analyzing text files.
Date/Time must be in the first data column. Its format can be given on the
command line, otherwise assumed %Y-%m-%d %H:%M:%S.''')
    hstr = 'Sort columns by name.'
    parser.add_argument('-c', '--columnsort', action='store_true',
                        default=csort, dest='csort', help=hstr)
    hstr  = "Format codes of the platform's C library strftime(),"
    hstr += " used by Python."
    parser.add_argument('-f', '--format', action='store', default=form,
                        dest='form', metavar='C-format', help=hstr)
    hstr  = ("header keyword of read_csv of pandas. --header='[0,1]' reads two"
             " line of headers. dfgui uses the first line for its column"
             " names")
    parser.add_argument('--header', action='store', default=header,
                        dest='header', metavar='read_csv_header', help=hstr)
    hstr = "Field delimiter (Default: try , ; None)"
    parser.add_argument('-s', '--sep', action='store', default=sep, dest='sep',
                        metavar='field_delimiter', help=hstr)
    hstr  = "Missing value. Will be set to NaN for plotting."
    hstr += " Use long version if negative number, e.g. --undef='-9999.'"
    parser.add_argument('-u', '--undef', action='store', type=float,
                        default=undef, dest='undef',
                        metavar='undef', help=hstr)
    hstr = 'Comma (,) or semi-colon (;) delimited text file(s).'
    parser.add_argument('files', nargs='*', default=None, metavar='file(s)',
                        help=hstr)

    args    = parser.parse_args(iargs)
    csort   = args.csort
    form    = args.form
    header  = eval(args.header)
    sep     = args.sep
    undef   = args.undef
    infiles = args.files

    if len(infiles) == 0:
        print('No input file given.')
        import sys
        sys.exit()

    del parser, args

    import datetime as dt
    import numpy as np
    import pandas as pd

    parser = lambda date: pd.to_datetime(date, format=form)

    infile = infiles[0]
    if sep is not None:
        df = pd.read_csv(infile, sep=sep, parse_dates=[0], date_parser=parser,
                         index_col=0, header=header)
    else:
        try:
            sep = ','
            df = pd.read_csv(infile, sep=sep, parse_dates=[0],
                             date_parser=parser, index_col=0, header=header)
        except:
            try:
                sep = ';'
                df = pd.read_csv(infile, sep=sep, parse_dates=[0],
                                 date_parser=parser, index_col=0,
                                 header=header)
            except:
                try:
                    # Pandas try to detect delimiter by using csv.Sniffer.
                    sep = None
                    df = pd.read_csv(infile, sep=sep, parse_dates=[0],
                                     date_parser=parser, index_col=0,
                                     header=header)
                except:
                    raise IOError('Pandas could not read input file: ' +
                                  infile)

    if len(infiles) > 1:
        for infile in infiles[1:]:
            df1 = pd.read_csv(infile, sep=sep, parse_dates=[0],
                              date_parser=parser, index_col=0, header=header)
            df = df.append(df1, sort=True)

    # no MultiIndex column names
    head = [ i[0] if not isinstance(i, str) else i for i in df.columns ]
    df.columns = head

    # # to concatenate files
    # df.to_csv('all.csv', na_rep=-9999., line_terminator='\r\n')

    if undef is not None:
        df.replace(undef, np.nan, inplace=True)

    if csort:
        df = df.reindex(sorted(df.columns), axis=1)

    # This must be before any other call to matplotlib
    # because it use the wxAgg backend.
    # This means, do not use --pylab with ipython.
    show(df)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def get_angle(xy1, xy2):
    """
        Returns the angle in radiant from each point in xy1 to each point in 
        xy2. Angles range from 0 (north) turning clockwise to 6.28... rad
        (360 degree).


        Definition
        ----------
        get_angle(xy1, xy2):


        Input
        -----
        xy1       2D np.array (n,2),
                  x and y coordinates of the points from where angles are desired 
        xy2       2D np.array (m,2), x and y coordinates of the points to where
                  angles are desired 


        Output
        ------
        angle     2D np.array (n,m), angles from each point in xy1 to all points
                  in xy2 in radiant.


        Restrictions
        ------------
        No support for 1D arrays.


        Example
        --------
        >>> ang = np.rad2deg(get_angle(np.array([[0.,0.]]), np.array([[0.,1.]])))
        >>> print(np.ma.round(ang).astype(int))
        [[0]]
        
        >>> ang = np.rad2deg(get_angle(np.array([[0.,0.]]), np.array([[1.,0.]])))
        >>> print(np.ma.round(ang).astype(int))
        [[90]]
        
        >>> ang = np.rad2deg(get_angle(np.array([[0.,0.]]), np.array([[-1.,0.]])))
        >>> print(np.ma.round(ang).astype(int))
        [[270]]
        
        >>> ang = np.rad2deg(get_angle(np.array([[0.,0.],[1.,1.]]), np.array([[1.,1.],[1.,0.]])))
        >>> print(np.ma.round(ang).astype(int))
        [[45 90]
         [-- 180]]
         

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Arndt Piayda

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Jun 2014
    """
    
    oldsettings = np.geterr()
    np.seterr(divide='ignore', invalid='ignore')
    
    rad180 = np.deg2rad(180.)
    rad360 = np.deg2rad(360.)
    angle = np.ma.empty((xy1.shape[0],xy2.shape[0]))
    for i, item in enumerate(xy1):
        adjacent_leg = xy2[:,1]-item[1]
        opposite_leg = xy2[:,0]-item[0]
        hypothenuse  = np.ma.sqrt((xy2[:,0]-item[0])**2 + (xy2[:,1]-item[1])**2)
        rad          = np.ma.arcsin(opposite_leg/hypothenuse)
        
        angle[i,:] = np.ma.where(
                                 # 1. quadrant
                                 (adjacent_leg >= 0.) & (opposite_leg >= 0.), rad,
                                 # 2. quadrant
                                 np.where((adjacent_leg <  0.) & (opposite_leg >= 0.), rad180-rad,
                                 # 3. quadrant
                                 np.where((adjacent_leg <  0.) & (opposite_leg <  0.), np.abs(rad)+rad180,
                                 # 4. quadrant
                                 rad360+rad)))
    
    np.seterr(**oldsettings)
    
    return angle

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
    Helper functions to deal with CHS level1 data and flags.


    Provided functions
    ------------------
    constant_values  Checks if a given series of data contains consecutive values which are constant over a certain time period.
    get_flag              Get the flags at position n from CHS data flag vector.
    get_manual_flags      Get start and end dates as well as flag values for a specific variable from a manual flag file.
    get_maxflag           Get the maximal flag of the string with the individual flags.
    get_header_excel      Get the header row of an Excel sheet.
    get_value_excel       Get value in column of sheet in excelfile given variable name.
    read_data             Read and concatenate data from CHS level1 data files.
    set_flag              Set the flags at position n to iflag at indices ii of CHS data flag vector.
    write_data            Write concatenated data back to individual CHS level1 data files.
    write_data_one_file   Write concatenated data back to one CHS data file.
    spike                 Spike filter for rectangularly shaped spikes.


    Example
    -------
    # get files
    infiles = jams.files_from_gui(title='Choose Level 1 file(s)')

    # read files
    sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead = jams.level1.read_data(infiles)

    # Get variables and indexes of variables in dat and flags
    myvars = [ v for v in hdat if v.startswith('stemT') ]
    idx    = [ hdat.index(v) for v in myvars ]

    # Set flags if variables were not treated yet
    flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

    # 1st test - check manual flags
    itest = 1
    jisdate = np.array(jams.date2dec(eng=isdate))
    for v in myvars:
        msdate, medate, mflags = get_manual_flags(manual_flag_file, v)
        if len(msdate) > 0:
            for dd in range(len(msdate)):
                ii = np.where((jsdate>=msdate[dd]) & (jsdate<=medate[dd]))[0]
                if ii.size>0: flags[:,i] = jams.level1.set_flag(flags[:,i], itest, mflags[dd], ii)

    # 2nd test - set first flag after the initial 9 to 2 if dat is undef
    itest  = 2
    isflag = 2
    for i in idx:
        ii = np.where(dat[:,i]==undef)[0]
        if ii.size>0: flags[:,i] = jams.level1.set_flag(flags[:,i], itest, isflag, ii)


    # 3rd test - set second flag to 2 if dat is not in [min,max]
    #            treat only data that had no flag==2 before
    itest = 3
    isflag = 2
    for i, v in zip(idx, myvars):
        mini = jams.level1.get_value_excel('CHS-measurements.xlsx', 'Forest Hohes Holz', v, 'Min')
        maxi = jams.level1.get_value_excel('CHS-measurements.xlsx', 'Forest Hohes Holz', v, 'Max')
        maxflags = jams.level1.get_maxflag(flags[:,i])
        ii = np.where((maxflags < 2) & ((dat[:,i] < swdrmin) | (dat[:,i] > swdrmax)))[0]
        if ii.size>0: flags[:,i] = jams.level1.set_flag(flags[:,i], itest, isflag, ii)

    # write back data and flags
    jams.level1.write_data(infiles, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead)

    # plot data without any flag
    i = 0
    xx = jams.date2dec(eng=sdate)
    yy = np.ma.array(dat[:,i], mask=(jams.level1.get_maxflag(flags[:,i])==2))
    mark1 = sub.plot(xx, yy)

    # Make level2
    flags = jams.maximum(jams.level1.get_maxflag(flags), 0)

    # write level2 data
    ofile = infiles[0].replace('level1', 'level2')
    jams.level1.write_data_onefile(ofile, sdate, record, dat, flags, hdate, hrecord, hdat, hflags)

    # write level2b data for DMP
    ofile = infiles[0].replace('level1', 'level2b')
    hdmp = jams.level1.get_value_excel(chsxlsfile, sheet, hdat, 'headerout (DB)')
    jams.level1.write_data_dmp(ofile, sdate, record, dat, flags, hdate, hrecord, hdat, hflags, hdmp)


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2015-2016 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  MC, Mar 2015
    Modified  JM, May 2015 - get_maxflag
              MC, May 2015 - excel - get_value_excel
              MC, Aug 2015 - get_manual_flags
              MC, Jan 2016 - get_header_excel
              MC, Mar 2016 - write_data_dmp, write_data_one_file
              BD, Sep 2016 - spike
"""
from .constant_values import constant_values
try:
    from .excel       import get_header_excel, get_value_excel
except:
    pass # No xlrd installed
from .getset_flag     import get_flag, set_flag, get_maxflag
from .manual_flags    import get_manual_flags
from .spike    import spike
from .readwrite_data  import read_data, write_data, write_data_dmp, write_data_dmp_size, write_data_one_file

# Information
__author__   = "Matthias Cuntz"
__version__  = '1.3'
__revision__ = "Revision: 2477"
__date__     = 'Date: 14.01.2016'



#!/usr/bin/env python
import numpy as np
from pyjams.mad import mad
from pyjams.jams.date2dec import date2dec


def spikeflag(date, data, inflag, isday, outdir, window=13, iter=1,
              fill_days=1, t_int=48, z=7, deriv=0, udef=-9999, spike_v=2,
              plot=False):
    '''
    Spike detection for Eddy Covariance data (and basically all other data)
    using a moving median absolute difference filter. Multiple iterations
    possible. Originally coded by Tino Rau.


    Definition
    ----------
    spikeflag(date, data, inflag, isday, window=13, iter=1,
              fill_days=1, t_int=48, z=5.5, deriv=0, udef=-9999, spike_v=2,
              plot=False):


    Input
    -----
    date        np.array(N), julian date (used only for plotting)
    data        np.array(N,M), data array where spike detection is applied on
                each column (M)
    inflag      np.array(N,M), dtype=int, quality flag of data, spike detection
                is only applied where inflag=0, all other data is ignored
    isday       np.array(N), dtype=bool, True where it is day and False where
                it is night
    outdir      path where plots are saved


    Optional Input
    --------------
    window      int, size of the moving window where mad is calculated in days
                (default: 13)
    iter        int, how often the running window mad shall be applied
                (default: 1)
    fill_days   int, number of days where mad is applied within moving window
                (default: 1)
    t_int       int, number of data points within one day (default: 48)
    z           int/float, data is allowed to deviate maximum z standard
                deviations from the median (default: 7)
    deriv       int, 0: Act on raw data; 1: use first derivatives;
                2: use 2nd derivatives (default: 0)
    udef        int/float, missing value of data (default: -9999) NaN values are
                excluded from computations anyhow.
    spike_v     int, spike value which shall be returned when a spike is
                detected (default: 2)
    plot        bool, if True data and spikes are plotted (default: False)


    Output
    ------
    flag        np.array(N), flag array where everything is 0 except where
                spikes were detected, there it is spike_v.


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Aug 2014
    '''
    rows, cols = np.shape(data)
    flag       = np.zeros_like(inflag).astype(np.int)
    # mad window length and flag window length
    period   = np.int(window*t_int)/2
    fill_win = np.int(fill_days*t_int)/2

    # calculate dusk and dawn times and separate in day and night
    isdawn      = np.zeros(rows,dtype=np.bool)
    isdusk      = np.zeros(rows,dtype=np.bool)
    dis         = (isday.astype(int) - np.roll(isday,-1).astype(int)).astype(bool)
    isdawn[:-1] = np.where(dis[:-1] == -1, True, False)
    isdusk[:-1] = np.where(dis[:-1] == 1, True, False)
    isddday     = isdawn
    tmp         = np.roll(isdusk,1)
    isddday[1:] += tmp[1:]
    isddnight   = isdusk
    tmp         = np.roll(isdawn,1)
    isddnight[1:] += tmp[1:]

    # iterate over each column of data
    for col in xrange(cols):
        # iterate as much as iter
        for i in xrange(iter):
            # get day and night data#
            day_data   = np.where((isday | isddday) & (inflag[:,col]==0) &
                                  ((data[:,col]!=udef) | (~np.isnan(data[:,col]))),
                                  data[:,col], np.nan)
            night_data = np.where((~isday | isddnight) & (inflag[:,col]==0) &
                                  ((data[:,col]!=udef) | (~np.isnan(data[:,col]))),
                                  data[:,col], np.nan)

            # iterate over flag window
            fill_points = xrange(fill_win, isday.size-1, 2*fill_win)
            for j in fill_points:
                j1 = np.max([ j - period - 1,0])
                j2 = np.min([ j + period + 1,isday.size])
                fill_start = np.max([ j - fill_win,1])
                fill_end   = np.min([ j + fill_win,isday.size-1])

                day_flag = mad(np.ma.masked_array(data=day_data[j1:j2],
                                                  mask=(np.isnan(day_data[j1:j2]))),
                               z=z, deriv=deriv)

                flag[fill_start:fill_end,col] += np.where(day_flag[fill_start-j1-1:fill_end-j1-1],
                                                          spike_v, 0)

                night_flag = mad(np.ma.masked_array(data=night_data[j1:j2],
                                                    mask=(np.isnan(night_data[j1:j2]))),
                                 z=z, deriv=deriv)

                flag[fill_start:fill_end,col] += np.where(night_flag[fill_start-j1-1:fill_end-j1-1],
                                                          spike_v, 0)

            if plot:
                import matplotlib as mpl
                import matplotlib.pyplot as plt
                import matplotlib.backends.backend_pdf as pdf
                majticks = mpl.dates.MonthLocator(bymonthday=1)
                format_str='%d %m %Y %H:%M'
                date01 = date2dec(yr=1, mo=1, dy=2, hr=0, mi=0, sc=0)

                fig1 = plt.figure(1)
                sub1 = fig1.add_subplot(111)
                valid = (inflag[:,col]==0) & ((data[:,col]!=udef) |
                                              (~np.isnan(data[:,col])))
                l1 =sub1.plot(date[valid]-date01, data[valid,col], '-b')
                l2 =sub1.plot(date[flag[:,col]!=0]-date01, data[flag[:,col]!=0,col], 'or')
                sub1.xaxis.set_major_locator(majticks)
                sub1.xaxis.set_major_formatter(mpl.dates.DateFormatter(format_str))
                fig1.autofmt_xdate()
                plt.show()

                pp1 = pdf.PdfPages(outdir+'/spike_%i.pdf'%col)
                fig1.savefig(pp1, format='pdf')
                pp1.close()

    return flag

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def maskgroup(x, n):
    """
        Masks elements in masked 1d-array gathered in groups of n elements.
        Retains any existing mask.

        For example for n=1: if an unmasked value is surrounded by two masked values, it will be masked as well.
        Or for n=2: two consecutive unmasked elements surrounded by masked elements will be masked.


        Definition
        ----------
        def maskgroup(x, n):


        Input
        -----
        x            1d-array
        n            int: any number <= n of consecutive unmasked data points surrounded by masked elements
                          should get masked


        Output
        ------
        new_mask     mask of x where additionally to the originally masked
                     elements, all elements are masked that are in a consecutive
                     group of data points with a length smaller equal to n


        Examples
        --------
        >>> # Create some data
        >>> x=np.ma.array([1,2,3,4,5,6,7,8,9,10,11,12], mask=[1,0,0,0,0,1,0,0,0,1,0,0])

        >>> # mask all elements which are gathered in groups smaller then 3
        >>> new_mask = maskgroup(x,2)
        >>> # apply new mask to x
        >>> print(np.ma.array(x.data, mask=new_mask))
        [-- 2 3 4 5 -- 7 8 9 -- -- --]

        >>> # mask all elements with are gathered in groups smaller equal to 3
        >>> new_mask = maskgroup(x,3)
        >>> # apply new mask to x
        >>> print(np.ma.array(x.data, mask=new_mask))
        [-- 2 3 4 5 -- -- -- -- -- -- --]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Feb 2014
        Modified, MC, Feb 2014 - call it maskgroup instead of small_kickout1d, mask <=n instead of <n
                  MC, Apr 2014 - removed enumerate and zip for simplification
    """
    # create running lists
    index  = []
    length = []
    count  = 0

    # count indices and lengths of unmasked data groups
    for i in range(x.mask.size):
        if i==0:
            if not x.mask[i]:
                index += [i]
                count  = 1
        if i>0:
            if not x.mask[i] and x.mask[i-1]:
                index += [i]
                count  = 1
            elif not x.mask[i]:
                count += 1
            elif x.mask[i] and not x.mask[i-1]:
                length += [count]
                count = 0
            else:
                pass
    if count>0:
        length += [count]

    # create new mask
    new_mask = np.copy(x.mask)
    for i in range(len(index)):
        if length[i]<=n:
            new_mask[index[i]:index[i]+length[i]] = True

    return new_mask

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""
This is the unittest for argsort module.

python -m unittest -v tests/test_argsort.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_argsort.py

"""
import unittest


class TestArgsort(unittest.TestCase):
    """
    Tests for argsort.py
    """

    def test_argmax(self):
        import numpy as np
        import pandas as pd
        from pyjams import argmax

        lst = [0, 4, 6, 2, 1, 5, 3, 5]

        # one-dimensional array
        a = np.array(lst)
        assert argmax(a) == 2
        assert a[argmax(a)] == 6

        # one-dimensional masked array
        a = np.ma.array(lst, mask=[0, 0, 1, 1, 0, 0, 0, 0])
        assert argmax(a) == 5
        assert a[argmax(a)] == 5
        assert argmax(a, fill_value=6) == 2

        # list
        a = lst[:]
        assert argmax(a) == 2
        assert a[argmax(a)] == 6

        # pandas
        df = pd.Series(lst)
        df.index = pd.date_range('2022-01-01', periods=len(lst))
        assert argmax(df) == 2
        assert df.iloc[argmax(df)] == 6

        # from numpy.argmax docstring
        a = np.arange(6).reshape(2, 3) + 10
        assert argmax(a) == 5
        self.assertEqual(list(argmax(a, axis=0)), [1, 1, 1])
        self.assertEqual(list(argmax(a, axis=1)), [2, 2])

        # indexes of the maximal elements of a n-dimensional array
        ind = np.unravel_index(argmax(a, axis=None), a.shape)
        self.assertEqual(list(ind), [1, 2])
        assert a[ind] == 15

        # Only the first occurrence is returned.
        b = np.arange(6)
        b[1] = 5
        assert argmax(b) == 1

    def test_argmin(self):
        import numpy as np
        import pandas as pd
        from pyjams import argmin

        lst = [0, 4, 6, 2, 1, 5, 3, 5]

        # one-dimensional array
        a = np.array(lst)
        assert argmin(a) == 0
        assert a[argmin(a)] == 0

        # one-dimensional masked array
        a = np.ma.array(lst, mask=[1, 0, 1, 1, 0, 0, 0, 0])
        assert argmin(a) == 4
        assert a[argmin(a)] == 1
        assert argmin(a, fill_value=1) == 0

        # list
        a = lst[:]
        assert argmin(a) == 0
        assert a[argmin(a)] == 0

        # pandas
        df = pd.Series(lst)
        df.index = pd.date_range('2022-01-01', periods=len(lst))
        assert argmin(df) == 0
        assert df.iloc[argmin(df)] == 0

        # from numpy.argmin docstring
        a = np.arange(6).reshape(2, 3) + 10
        assert argmin(a) == 0
        self.assertEqual(list(argmin(a, axis=0)), [0, 0, 0])
        self.assertEqual(list(argmin(a, axis=1)), [0, 0])

        # indexes of the maximal elements of a n-dimensional array
        ind = np.unravel_index(argmin(a, axis=None), a.shape)
        self.assertEqual(list(ind), [0, 0])
        assert a[ind] == 10

        # Only the first occurrence is returned.
        b = np.arange(6) + 10
        b[4] = 10
        assert argmin(b) == 0

    def test_argsort(self):
        import numpy as np
        from numpy.ma import masked
        import pandas as pd
        from pyjams import argsort

        lst = [0, 4, 6, 2, 1, 5, 3, 5]
        slst = [0, 1, 2, 3, 4, 5, 5, 6]

        # 1D array
        a = np.array(lst)
        self.assertEqual(list(a[argsort(a)]), slst)

        # sort algorithm
        self.assertEqual(list(a[argsort(a, kind='quicksort')]),
                         slst)

        # 1D masked array
        a = np.ma.array(lst, mask=[0, 0, 1, 1, 0, 0, 0, 0])
        self.assertEqual(list(a[argsort(a)]),
                         [0, 1, 3, 4, 5, 5, masked, masked])
        self.assertEqual(list(a[argsort(a, fill_value=1)]),
                         [0, masked, masked, 1, 3, 4, 5, 5])

        # list
        a = lst[:]
        ii = argsort(a)
        b = [ a[i] for i in ii ]
        self.assertEqual(b, slst)
        self.assertRaises(KeyError, argsort, a, key=a)

        a = lst[:]
        ii = argsort(a, reverse=True)
        b = [ a[i] for i in ii ]
        self.assertEqual(b, slst[::-1])

        # pandas
        df = pd.Series(lst)
        df.index = pd.date_range('2022-01-01', periods=len(lst))
        self.assertEqual(list(df.iloc[argsort(df)]), slst)

        # from numpy.argsort docstring
        # one-dimensional array
        x = np.array([3, 1, 2])
        self.assertEqual(list(argsort(x)), [1, 2, 0])

        # two-dimensional array
        x = np.array([[0, 3], [2, 2]])
        # sorts along first axis (down)
        ind = argsort(x, axis=0)
        # self.assertEqual(list(ind), [[0, 1], [1, 0]])
        self.assertEqual(list(ind.flatten()), [0, 1, 1, 0])
        # same as np.sort(x, axis=0)
        ind = np.take_along_axis(x, ind, axis=0)
        self.assertEqual(list(ind.flatten()), [0, 2, 2, 3])
        # sorts along last axis (across)
        ind = argsort(x, axis=1)
        self.assertEqual(list(ind.flatten()), [0, 1, 0, 1])
        # same as np.sort(x, axis=1)
        ind = np.take_along_axis(x, ind, axis=1)
        self.assertEqual(list(ind.flatten()), [0, 3, 2, 2])

        # indices of the sorted elements of a N-dimensional array
        ind = np.unravel_index(argsort(x, axis=None), x.shape)
        ind1 = np.array([ list(i) for i in ind ])
        self.assertEqual(list(ind1.flatten()), [0, 1, 1, 0, 0, 0, 1, 1])
        # same as np.sort(x, axis=None)
        self.assertEqual(list(x[ind]), [0, 2, 2, 3])

        # sorting with keys
        x = np.array([(1, 0), (0, 1)], dtype=[('x', '<i4'), ('y', '<i4')])
        self.assertEqual(list(argsort(x, order=('x', 'y'))), [1, 0])
        self.assertEqual(list(argsort(x, order=('y', 'x'))), [0, 1])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def climate_index_knoben(time, precip, tave, pet, snow=None, color=True, indicators=False):
    """
    Calculates climate index (color and/or indicators) based on Knoben et al. (2018).

    Addition: The fraction of precipitation as snow is based on snow time series given rather 
    than using Eq. 3 of Knoben et al. (2018).

    The daily precipitation, avergage daily temperature, and potential evapotranspiration 
    is converted into climate indicators aridity I_m, Seasonality I_m,r, and Precipitation 
    as snow f_S. These three are used to derive a color which is used to identify the 
    continuous climatic zones.

    Knoben, W. J. M., Woods, R. A., & Freer, J. E. (2018). 
    A Quantitative Hydrological Climate Classification Evaluated With Independent Streamflow Data. 
    Water Resources Research, 54(7), 5088-5109. 
    http://doi.org/10.1029/2018WR022913


    Definition
    ----------
    climate_index_knoben(time, precip, tave, pet, snow=None, color=True, indicators=False):


    Input
    -----
    time         array of datetime objects of time stamps
    precip       array of daily precipitation (snow+rain)
    tave         array of daily average temperature
    pet          array of daily potential evapotranspiration


    Options
    -------
    snow         array of daily average snow (uses this estimate rather 
                 than Eq. 3 of Knoben et al. (2018).
    color        triplet of RGB values for 
                 - aridity (red), 
                 - seasonality (green), and 
                 - precipitation as snow (blue) (range: [0,1]) 
    indictators  triplet of indicator values for 
                 - aridity (arid to wet) [-1,1], 
                 - seasonality (constant to seasonal) [0,2], and 
                 - precipitation as snow (no snow to all snow) [0,1] 

    Output
    ------
    Dictionary of outputs selected:
    color = True:
       {color: {red: <red>, green: <green>, blue: <blue>}}
    indicators = True:
       {indicators: {aridity: <aridity>, seasonality: <seasonality>, precip_as_snow: <precip_as_snow>}}


    Restrictions
    ------------
    None.


    Examples
    --------
    # Climate index (color and/or indicators) based on Knoben et al. (2018).
    >>> import numpy as np
    >>> import datetime as datetime
    >>>
    >>> np.random.seed(1)
    >>> time   = np.array([ datetime.datetime(1950,1,1,0,0)+datetime.timedelta(itime) for itime in range(1000) ])
    >>> precip = np.random.random(1000)*20   
    >>> tave   = np.random.random(1000)*20-10.   
    >>> pet    = np.random.random(1000)*100   
    >>>
    >>> climate_index = climate_index_knoben(time, precip, tave, pet, color=True, indicators=True)
    >>>
    >>> from autostring import astr
    >>> print(astr([climate_index['color']['red'],climate_index['color']['green'],climate_index['color']['blue']],3,pp=True))
    ['0.896' '0.039' '0.165']
    >>> print(astr([climate_index['indicators']['aridity'],climate_index['indicators']['seasonality'],climate_index['indicators']['precip_as_snow']],3,pp=True))
    ['-0.792' ' 0.077' ' 0.165']


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2020 Juliane Mai - juliane.mai@uwaterloo.ca 

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  JM, Feb 2020
    """

    import numpy  as np
    import pandas as pd

    # put all in dataframe to make averaging easier
    data = np.transpose( np.array([ precip, tave, pet]) )
    df = pd.DataFrame(data, columns = ['precip', 'tave', 'pet'], index=pd.DatetimeIndex(time), dtype=float)

    # monthly sum of precip
    precip_total_monthly     = df.precip.resample("M").agg(['sum'])                                  # 12 x 61 values
    precip_total_monthly_ave = precip_total_monthly.groupby(precip_total_monthly.index.month).mean() # 12 values

    # average monthly temperature
    tave_ave_monthly     = df.tave.resample("M").agg(['mean'])                                       # 12 x 61 values
    tave_ave_monthly_ave = tave_ave_monthly.groupby(tave_ave_monthly.index.month).mean()             # 12 values

    # average monthly pet
    pet_total_monthly     = df.pet.resample("M").agg(['sum'])                                        # 12 x 61 values
    pet_total_monthly_ave = pet_total_monthly.groupby(pet_total_monthly.index.month).mean()          # 12 values

    # >>>>>>>>>>>>>> TODO: write these three to file

    # Equation 1 in Knoben et al. (2018): moisture index MI
    nmonths = 12
    MI = np.zeros(nmonths)
    for itime in range(nmonths):
        if precip_total_monthly_ave.values[itime] > pet_total_monthly_ave.values[itime]:
            MI[itime] = 1.0 - pet_total_monthly_ave.values[itime] / precip_total_monthly_ave.values[itime]
        elif precip_total_monthly_ave.values[itime] < pet_total_monthly_ave.values[itime]:
            MI[itime] = precip_total_monthly_ave.values[itime] / pet_total_monthly_ave.values[itime] - 1.0
        else:
            MI[itime] = 0.0

    # Equation 2 in Knoben et al. (2018)
    #  ---> Red   = Aridity I_m
    #       -1 = arid
    #        1 = wet
    I_m = np.mean(MI)

    # Equation 3 in Knoben et al. (2018)
    #  ---> Green = Seasonality I_m,r
    #       0 = constant
    #       2 = seasonal
    I_mr = np.max(MI) - np.min(MI)

    if (snow is None):
        # Equation 4 in Knoben et al. (2018)
        #  ---> Blue  = Fraction of precipitation as snow f_S
        #       0 = no snow
        #       1 = all snow
        f_S = np.sum( precip_total_monthly_ave.values[tave_ave_monthly_ave.values <= 0.0] ) / np.sum( precip_total_monthly_ave.values )
    else:
        # if snow farction is given, ratio of total snow amount and total precipitation amount is used to derive snow fraction
        f_S = np.sum( snow ) / np.sum( precip_total_monthly.values )
        # print('snow amount: ',np.sum( snow ))
        # print('prec amount: ',np.sum( precip_total_monthly.values ))

    # print("Aridity I_m                           = ", I_m)
    # print("Seasonality I_m,r                     = ", I_mr)
    # print("Fraction of precipitation as snow f_S = ", f_S)


    # color mapping
    # R = 1 - (aridity_Im + 1)./2       --> transforms [-1,1] to [1,0] --> [most arid, least arid]
    # G = (aridity_seasonality_Imr)./2  --> transforms [ 0,2] to [0,1] --> [least seasonal, most seasonal]
    # B = precipitation_as_snow_fs      --> transforms [ 0,1] to [0,1] --> [least snow, most snow]
    red   = 1.0 - (I_m + 1.0)/ 2.0
    green = I_mr/ 2.0
    blue  = f_S

    # print("Red   = ", red)
    # print("Green = ", green)
    # print("Blue  = ", blue)

    output = {}
    if color:
        output['color'] = {'red': red, 'green': green, 'blue': blue}
    if indicators:
        output['indicators'] = { 'aridity': I_m, 'seasonality': I_mr, 'precip_as_snow': f_S}

    return output

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
import scipy.optimize as opt  # curve_fit, fmin, fmin_tnc
import pyjams.functions as functions  # from jams
from pyjams.mad import mad # from jams
import warnings
# import pdb


# ----------------------------------------------------------------------
def nee2gpp(dates, nee, t, isday, rg=False, vpd=False, undef=np.nan,
            method='reichstein', shape=False, masked=False, nogppnight=False):
    """
        Calculate photosynthesis (GPP) and ecosystem respiration (Reco) from original
        Eddy flux data.

        It uses either
          1. a fit of Reco vs. temperature to all nighttime data, or
          2. several fits over the season of Reco vs. temperature as in Reichstein et al. (2005), or
          3. the daytime method of Lasslop et al. (2010),
        in order to calculate Reco and then GPP = Reco - NEE.


        Definition
        ----------
        def nee2gpp(dates, nee, t, isday, rg=False, vpd=False, undef=np.nan,
                    method='reichstein', shape=False, masked=False):


        Input
        -----
        Inputs are 1D arrays that can be masked or not.
        dates         julian days
        nee           net ecosystem exchange (uptake is <0) [umol m-2 s-1]
        t             temperature [K]


        Optional Input
        --------------
        If method = 'day' | 'lasslop', extra inputs are
        rg            global radiation, i.e. shortwave down [W m-2]
        vpd           vapour pressure deficit [Pa]


        Parameters
        ----------
        undef        undefined values in data  (default: np.nan)
                     Input arrays will be masked at undef, keeping the original mask
        method       if 'global' | 'falge':      fit of Reco vs. temperature to all nighttime data
                     if 'local'  | 'reichstein': method of Reichstein et al. (2005)
                     if 'day'    | 'lasslop':    method of Lasslop et al. (2010)
        shape        if False then outputs are 1D arrays;
                     if True, output have the same shape as datain
                     if a shape tuple is given, then this tuple is used to reshape
        masked       if False: outputs are undef where nee and t are masked or undef
                     if True:  return masked arrays where outputs would be undef

        If method = 'night' | 'reichstein', extra parameters are
        nogppnight   if True:  Resp=NEE, GPP=0 at night, GPP always positive
                     if False: Resp=lloyd_taylor, GPP=Resp-NEE at night (default)


        Ouput
        -----
        GPP, Reco    photosynthesis, ecosystem respiration


        Restrictions
        ------------
        Negative respiration possible at night when gpp is forced to 0 with nogppnight=True


        Literature
        ----------
        Falge et al. (2001)
            Gap filling strategies for defensible annual sums of net ecosystem exchange
            Acricultural and Forest Meteorology 107, 43-69

        Lasslop et al. (2010)
            Separation of net ecosystem exchange into assimilation and respiration using
            a light response curve approach: critical issues and global evaluation
            Global Change Biology 16, 187-208

        Reichstein et al. (2005)
            On the separation of net ecosystem exchange into assimilation and ecosystem
            respiration: review and improved algorithm.
            Global Change Biology 11, 1424-1439


        Examples
        --------
        >>> from jams.fread import fread # from jams
        >>> from jams.date2dec import date2dec # from jams
        >>> dat   = fread('test_nee2gpp.csv', skip=2, transpose=True)
        >>> dates = date2dec(dy=dat[0,:], mo=dat[1,:], yr=dat[2,:], hr=dat[3,:], mi=dat[4,:])
        >>> NEE   = np.squeeze(dat[5,:])
        >>> rg    = np.squeeze(dat[6,:])
        >>> tair  = np.squeeze(dat[7,:])
        >>> undef = -9999.
        >>> isday = np.where(rg > 10., True, False)
        >>> tt    = np.where(tair == undef, undef, tair+273.15)
        >>> # partition
        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='local')
        >>> print(GPP[1120:1128])
        [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.40606871e+00
          8.31942152e+00  1.06242542e+01  8.49245664e+00  1.12381973e+01]
        >>> print(Reco[1120:1128])
        [1.68311981 1.81012431 1.9874173  2.17108871 2.38759152 2.64372415
         2.90076664 3.18592735]

        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='local')
        >>> print(GPP[1120:1128])
        [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.40606871e+00
          8.31942152e+00  1.06242542e+01  8.49245664e+00  1.12381973e+01]

        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='global')
        >>> print(GPP[1120:1128])
        [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.33166157e+00
          8.18228013e+00  1.04092252e+01  8.19395317e+00  1.08427448e+01]

        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='Reichstein', masked=True)
        >>> print(GPP[1120:1128])
        [-- -- -- 4.406068706013192 8.319421516040766 10.624254150217764
         8.492456637225963 11.238197347837367]

        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='reichstein', shape=(np.size(NEE),1))
        >>> print(GPP[1120:1128])
        [[-9.99900000e+03]
         [-9.99900000e+03]
         [-9.99900000e+03]
         [ 4.40606871e+00]
         [ 8.31942152e+00]
         [ 1.06242542e+01]
         [ 8.49245664e+00]
         [ 1.12381973e+01]]

        >>> VPD = np.squeeze(dat[8,:])
        >>> vpd = np.where(VPD == undef, undef, VPD*100.)
        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, rg, vpd, undef=undef, method='day')
        >>> print(GPP[1120:1128])
        [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  2.78457540e+00
          6.63212545e+00  8.88902165e+00  6.74243873e+00  9.51364527e+00]
        >>> print(Reco[1120:1128])
        [0.28786696 0.34594516 0.43893276 0.5495954  0.70029545 0.90849165
         1.15074873 1.46137527]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2014 Matthias Cuntz, Arndt Piayda - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written  MC, Mar 2012
        Modified AP, Mar 2012 - undef=np.nan
                 MC, Nov 2012 - wrapper for individual routines nee2gpp_reichstein etc.
                 MC, Feb 2013 - ported to Python 3
                 MC, May 2013 - replaced cost functions by generel cost function cost_abs if possible
                 AP, Aug 2014 - replaced fmin with fmin_tnc to permit params<0,
                                permit gpp<0 at any time if nogppnight=True
    """

    # Global relationship in Reichstein et al. (2005)
    if ((method.lower() == 'global') | (method.lower() == 'falge')):
        return nee2gpp_falge(dates, nee, t, isday, undef=undef, shape=shape, masked=masked)
    # Local relationship = Reichstein et al. (2005)
    elif ((method.lower() == 'local') | (method.lower() == 'reichstein')):
        return nee2gpp_reichstein(dates, nee, t, isday, undef=undef, shape=shape, masked=masked, nogppnight=nogppnight)
    # Lasslop et al. (2010) method
    elif ((method.lower() == 'day') | (method.lower() == 'lasslop')):
        return nee2gpp_lasslop(dates, nee, t, isday, rg, vpd, undef=undef, shape=shape, masked=masked, nogppnight=nogppnight)
    # Include new methods here
    else:
        raise ValueError('Error nee2gpp: method not implemented yet.')



# ----------------------------------------------------------------------
def nee2gpp_falge(dates, nee, t, isday, undef=np.nan,
            shape=False, masked=False):
    """
        Calculate photosynthesis (GPP) and ecosystem respiration (Reco) from original
        Eddy flux data, using a fit of Reco vs. temperature to all nighttime data,
        in order to calculate Reco and then GPP = Reco - NEE.


        Definition
        ----------
        def nee2gpp_falge(dates, nee, t, isday, undef=np.nan, shape=False, masked=False):


        Input
        -----
        Inputs are 1D arrays that can be masked or not.
        dates         julian days
        nee           net ecosystem exchange (uptake is <0) [umol m-2 s-1]
        t             temperature [K]


        Parameters
        ----------
        undef        undefined values in data  (default: np.nan)
                     Input arrays will be masked at undef, keeping the original mask
        shape        if False then outputs are 1D arrays;
                     if True, output have the same shape as datain
                     if a shape tuple is given, then this tuple is used to reshape
        masked       if False: outputs are undef where nee and t are masked or undef
                     if True:  return masked arrays where outputs would be undef


        Ouput
        -----
        GPP, Reco    photosynthesis, ecosystem respiration


        Restrictions
        ------------
        None.


        Literature
        ----------
        Falge et al. (2001)
            Gap filling strategies for defensible annual sums of net ecosystem exchange
            Acricultural and Forest Meteorology 107, 43-69


        Examples
        --------
        >>> from jams.fread import fread # from jams
        >>> from jams.date2dec import date2dec # from jams
        >>> dat   = fread('test_nee2gpp.csv', skip=2, transpose=True)
        >>> dates = date2dec(dy=dat[0,:], mo=dat[1,:], yr=dat[2,:], hr=dat[3,:], mi=dat[4,:])
        >>> NEE   = np.squeeze(dat[5,:])
        >>> rg    = np.squeeze(dat[6,:])
        >>> tair  = np.squeeze(dat[7,:])
        >>> undef = -9999.
        >>> isday = np.where(rg > 10., True, False)
        >>> tt    = np.where(tair == undef, undef, tair+273.15)
        >>> # partition
        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='global')
        >>> print(GPP[1120:1128])
        [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.33166157e+00
          8.18228013e+00  1.04092252e+01  8.19395317e+00  1.08427448e+01]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2013 Matthias Cuntz, Arndt Piayda - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written  MC, Mar 2012
        Modified AP, Mar 2012 - undef=np.nan
                 MC, Nov 2012 - individual routine
                 MC, Feb 2013 - ported to Python 3
    """

    # Checks

    # remember shape if any
    inshape = nee.shape
    dates   = np.squeeze(dates)
    nee     = np.squeeze(nee)
    t       = np.squeeze(t)
    isday   = np.squeeze(isday)
    # Check squeezed shape
    if dates.ndim != 1: raise Error('Error nee2gpp_falge: squeezed dates must be 1D array.')
    if nee.ndim   != 1: raise Error('Error nee2gpp_falge: squeezed nee must be 1D array.')
    if t.ndim     != 1: raise Error('Error nee2gpp_falge: squeezed t must be 1D array.')
    if isday.ndim != 1: raise Error('Error nee2gpp_falge: squeezed isday must be 1D array.')
    ndata = dates.size
    if ((nee.size != ndata) | (t.size != ndata) | (isday.size != ndata)):
        raise Error('Error nee2gpp_falge: inputs must have the same size.')

    # Transform to masked array with 1D mask
    nee   = np.ma.array(nee, mask=False)
    t     = np.ma.array(t, mask=False)
    isday = np.ma.array(isday, mask=False)
    # mask also undef
    if np.isnan(undef):
        if np.ma.any(np.isnan(nee)):   nee[np.isnan(nee)]     = np.ma.masked
        if np.ma.any(np.isnan(t)):     t[np.isnan(t)]         = np.ma.masked
        if np.ma.any(np.isnan(isday)): isday[np.isnan(isday)] = np.ma.masked
    else:
    	if np.ma.any(nee==undef):   nee[nee==undef]     = np.ma.masked
    	if np.ma.any(t==undef):     t[t==undef]         = np.ma.masked
    	if np.ma.any(isday==undef): isday[isday==undef] = np.ma.masked

    # Partition - Global relationship as in Falge et al. (2001)

    # Select valid nighttime
    mask = isday | nee.mask | t.mask | isday.mask
    ii   = np.where(~mask)[0]
    tt   = np.ma.compressed(t[ii])
    net  = np.ma.compressed(nee[ii])
    # p, c     = opt.curve_fit(functions.lloyd_fix, tt, net, p0=[2.,200.]) # global parameter, global cov matrix
    #p        = opt.fmin(functions.cost_lloyd_fix, [2.,200.], args=(tt, net), disp=False)
    p        = opt.fmin(functions.cost_abs, [2.,200.], args=(functions.lloyd_fix_p, tt, net), disp=False)

    Reco     = np.ones(ndata)*undef
    ii       = np.where(~t.mask)[0]
    Reco[ii] = functions.lloyd_fix(t[ii], p[0], p[1])

    # GPP
    GPP     = np.ones(ndata)*undef
    ii      = np.where(~(t.mask | nee.mask))[0]
    GPP[ii] = Reco[ii] - nee[ii]

    # Return
    if masked:
        if np.isnan(undef):
            GPP  = np.ma.array(GPP,  mask=np.isnan(GPP))
            Reco = np.ma.array(Reco, mask=np.isnan(Reco))
        else:
            GPP  = np.ma.array(GPP,  mask=(GPP == undef))
            Reco = np.ma.array(Reco, mask=(Reco == undef))

    if shape != False:
        if shape != True:
            return np.reshape(GPP,shape), np.reshape(Reco,shape)
        else:
            return np.reshape(GPP,inshape), np.reshape(Reco,inshape)
    else:
        return GPP, Reco



# ----------------------------------------------------------------------
def nee2gpp_reichstein(dates, nee, t, isday, rg=False, vpd=False, undef=np.nan,
            shape=False, masked=False, nogppnight=False):
    """
        Calculate photosynthesis (GPP) and ecosystem respiration (Reco) from original
        Eddy flux data, using several fits of Reco vs. temperature of nighttime data
        over the season, as in Reichstein et al. (2005), in order to calculate Reco
        and then GPP = Reco - NEE.


        Definition
        ----------
        def nee2gpp_reichstein(dates, nee, t, isday, undef=np.nan, shape=None, masked=False):


        Input
        -----
        Inputs are 1D arrays that can be masked or not.
        dates         julian days
        nee           net ecosystem exchange (uptake is <0) [umol m-2 s-1]
        t             temperature [K]


        Parameters
        ----------
        undef        undefined values in data  (default: np.nan)
                     Input arrays will be masked at undef, keeping the original mask
        shape        if False then outputs are 1D arrays (default)
                     if True, output have the same shape as datain
                     if a shape tuple is given, then this tuple is used to reshape
        masked       if False: outputs are undef where nee and t are masked or undef (default)
                     if True:  return masked arrays where outputs would be undef
        nogppnight   if True:  Resp=NEE, GPP=0 at night
                     if False: Resp=lloyd_taylor, GPP=Resp-NEE at night (default)


        Ouput
        -----
        GPP, Reco    photosynthesis, ecosystem respiration


        Restrictions
        ------------
        None.


        Literature
        ----------
        Reichstein et al. (2005)
            On the separation of net ecosystem exchange into assimilation and ecosystem
            respiration: review and improved algorithm.
            Global Change Biology 11, 1424-1439


        Examples
        --------
        >>> from jams.fread import fread # from jams
        >>> from jams.date2dec import date2dec # from jams
        >>> dat   = fread('test_nee2gpp.csv', skip=2, transpose=True)
        >>> dates = date2dec(dy=dat[0,:], mo=dat[1,:], yr=dat[2,:], hr=dat[3,:], mi=dat[4,:])
        >>> NEE   = np.squeeze(dat[5,:])
        >>> rg    = np.squeeze(dat[6,:])
        >>> tair  = np.squeeze(dat[7,:])
        >>> undef = -9999.
        >>> isday = np.where(rg > 10., True, False)
        >>> tt    = np.where(tair == undef, undef, tair+273.15)
        >>> # partition
        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='local')
        >>> print(GPP[1120:1128])
        [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.40606871e+00
          8.31942152e+00  1.06242542e+01  8.49245664e+00  1.12381973e+01]
        >>> print(Reco[1120:1128])
        [1.68311981 1.81012431 1.9874173  2.17108871 2.38759152 2.64372415
         2.90076664 3.18592735]

        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='local')
        >>> print(GPP[1120:1128])
        [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.40606871e+00
          8.31942152e+00  1.06242542e+01  8.49245664e+00  1.12381973e+01]

        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='Reichstein', masked=True)
        >>> print(GPP[1120:1128])
        [-- -- -- 4.406068706013192 8.319421516040766 10.624254150217764
         8.492456637225963 11.238197347837367]

        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='reichstein', shape=(np.size(NEE),1))
        >>> print(GPP[1120:1128])
        [[-9.99900000e+03]
         [-9.99900000e+03]
         [-9.99900000e+03]
         [ 4.40606871e+00]
         [ 8.31942152e+00]
         [ 1.06242542e+01]
         [ 8.49245664e+00]
         [ 1.12381973e+01]]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2013 Matthias Cuntz, Arndt Piayda - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written  MC, Mar 2012
        Modified AP, Mar 2012 - undef=np.nan
                 MC, Nov 2012 - individual routine
                 MC, Feb 2013 - ported to Python 3
    """
    # Checks

    # remember shape if any
    if shape != False:
        if shape != True:
            inshape = shape
        else:
            inshape = nee.shape
    dates   = np.squeeze(dates)
    nee     = np.squeeze(nee)
    t       = np.squeeze(t)
    isday   = np.squeeze(isday)
    if shape == False: inshape = nee.shape
    # Check squeezed shape
    if dates.ndim != 1: raise ValueError('Error nee2gpp_reichstein: squeezed dates must be 1D array.')
    if nee.ndim   != 1: raise ValueError('Error nee2gpp_reichstein: squeezed nee must be 1D array.')
    if t.ndim     != 1: raise ValueError('Error nee2gpp_reichstein: squeezed t must be 1D array.')
    if isday.ndim != 1: raise ValueError('Error nee2gpp_reichstein: squeezed isday must be 1D array.')
    ndata = dates.size
    if ((nee.size != ndata) | (t.size != ndata) | (isday.size != ndata)):
        raise ValueError('Error nee2gpp_reichstein: inputs must have the same size.')

    # Transform to masked array with 1D mask
    nee   = np.ma.array(nee, mask=False)
    t     = np.ma.array(t, mask=False)
    isday = np.ma.array(isday, mask=False)
    # mask also undef
    if np.isnan(undef):
        if np.ma.any(np.isnan(nee)):   nee[np.isnan(nee)]     = np.ma.masked
        if np.ma.any(np.isnan(t)):     t[np.isnan(t)]         = np.ma.masked
        if np.ma.any(np.isnan(isday)): isday[np.isnan(isday)] = np.ma.masked
    else:
    	if np.ma.any(nee==undef):   nee[nee==undef]     = np.ma.masked
    	if np.ma.any(t==undef):     t[t==undef]         = np.ma.masked
    	if np.ma.any(isday==undef): isday[isday==undef] = np.ma.masked

    # Partition - Local relationship = Reichstein et al. (2005)

    # Select valid nighttime
    mask = isday | nee.mask | t.mask | isday.mask
    ii   = np.where(~mask)[0]
    if (ii.size==0):
        print('Warning nee2gpp_reichstein: no valid nighttime data.')
        if masked:
            GPP  = np.ma.array(np.reshape(nee,inshape), mask=np.ones(inshape, dtype=bool))
            Reco = np.ma.array(np.reshape(nee,inshape), mask=np.ones(inshape, dtype=bool))
        else:
            GPP  = np.ones(np.reshape(nee,inshape))*undef
            Reco = np.ones(np.reshape(nee,inshape))*undef
        return GPP, Reco
    jul  = dates[ii]
    tt   = np.ma.compressed(t[ii])
    net  = np.ma.compressed(nee[ii])
    # 1. each 5 days, in 15 day period, fit if range of T > 5
    locp = [] # local param
    locs = [] # local err
    dmin = np.floor(np.amin(jul)).astype(int) # be aware that julian days starts at noon, i.e. 1.0 is 12h
    dmax = np.ceil(np.amax(jul)).astype(int)  # so the search will be from noon to noon and thus includes all nights
    for i in range(dmin,dmax,5):
        iii  = np.where((jul>=i) & (jul<(i+14)))[0]
        niii = iii.size
        if niii > 6:
            tt1  = tt[iii]
            net1 = net[iii]
            mm   = ~mad(net1, z=4.5) # make fit more robust by removing outliers
            if (np.ptp(tt[iii]) >= 5.) & (np.sum(mm) > 6):
                # print(i)
                #p     = opt.fmin(functions.cost_lloyd_fix, [2.,200.], args=(tt1[mm], net1[mm]), disp=False) # robust params

                p, temp1, temp2 = opt.fmin_tnc(functions.cost_lloyd_fix, [2.,200.], bounds=[[0.,None],[0.,None]],
                                              args=(tt1[mm], net1[mm]),
                                              approx_grad=True, disp=False)

                try:
                    p1, c = opt.curve_fit(functions.lloyd_fix, tt1[mm], net1[mm], p0=p, maxfev=10000) # params, covariance
                    if np.all(np.isfinite(c)): # possible return of curvefit: c=inf
                        s = np.sqrt(np.diag(c))
                    else:
                        s = 10.*np.abs(p)
                except:
                    s = 10.*np.abs(p)
                locp += [p]
                locs += [s]
                # if ((s[1]/p[1])<0.5) & (p[1] > 0.): pdb.set_trace()
    if len(locp) == 0:
        raise ValueError('Error nee2gpp_reichstein: No local relationship found.')
        print('Warning nee2gpp_reichstein: No local relationship found.')
        if masked:
            GPP  = np.ma.array(np.reshape(nee,inshape), mask=np.ones(inshape, dtype=bool))
            Reco = np.ma.array(np.reshape(nee,inshape), mask=np.ones(inshape, dtype=bool))
        else:
            GPP  = np.ones(np.reshape(nee,inshape))*undef
            Reco = np.ones(np.reshape(nee,inshape))*undef
        return GPP, Reco
    locp   = np.squeeze(np.array(locp).astype(float))
    locs   = np.squeeze(np.array(locs).astype(float))
    # 2. E0 = avg of best 3
    # Reichstein et al. (2005), p. 1430, 1st paragraph.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        iii  = np.where((locp[:,1] > 0.) & (locp[:,1] < 450.) & (np.abs(locs[:,1]/locp[:,1]) < 0.5))[0]
    niii = iii.size
    if niii==0:
        # raise ValueError('Error nee2gpp_reichstein: No good local relationship found.')
        # loosen the criteria: take the best three estimates anyway
        iii   = np.where((locp[:,1] > 0.))[0]
        niii = iii.size
        if niii<1:
            raise ValueError('Error nee2gpp_reichstein: No E0>0 found.')
            print('Warning nee2gpp_reichstein: No E0>0 found.')
            if masked:
                GPP  = np.ma.array(np.reshape(nee,inshape), mask=np.ones(inshape, dtype=bool))
                Reco = np.ma.array(np.reshape(nee,inshape), mask=np.ones(inshape, dtype=bool))
            else:
                GPP  = np.ones(np.reshape(nee,inshape))*undef
                Reco = np.ones(np.reshape(nee,inshape))*undef
            return GPP, Reco
        lp    = locp[iii,:]
        ls    = locs[iii,:]
        iis   = np.argsort(ls[:,1])
        bestp = np.mean(lp[iis[0:np.minimum(3,niii)],:],axis=0)
        bests = np.mean(ls[iis[0:np.minimum(3,niii)],:],axis=0)
    elif niii==1:
        bestp = np.squeeze(locp[iii,:])
        bests = np.squeeze(locs[iii,:])
    elif niii==2:
        bestp = np.mean(locp[iii,:],axis=0)
        bests = np.mean(locs[iii,:],axis=0)
        # ls    = locs[iii,:]
        # iis   = np.argsort(ls[:,1])
    else:
        lp    = locp[iii,:]
        ls    = locs[iii,:]
        iis   = np.argsort(ls[:,1])
        bestp = np.mean(lp[iis[0:3],:],axis=0)
        bests = np.mean(ls[iis[0:3],:],axis=0)

    # 3. Refit Rref with fixed E0, each 4 days
    refp  = [] # Rref param
    refii = [] # mean index of data points
    E0    = bestp[1]
    et    = functions.lloyd_fix(tt, 1., E0)
    for i in range(dmin,dmax,4):
        iii  = np.where((jul>=i) & (jul<(i+4)))[0]
        niii = iii.size
        if niii > 3:
            # Calc directly minisation of (nee-p*et)**2
            # p = np.sum(net[iii]*et[iii])/np.sum(et[iii]**2)
            # p, c = opt.curve_fit(functions.lloyd_only_rref, et[iii], net[iii], p0=[2.])
            #p      = opt.fmin(functions.cost_lloyd_only_rref, [2.], args=(et[iii], net[iii]), disp=False)
            #p = opt.fmin(functions.cost_abs, [2.], args=(functions.lloyd_only_rref_p, et[iii], net[iii]), disp=False)

            p, temp1, temp2 = opt.fmin_tnc(functions.cost_abs, [2.], bounds=[[0.,None]],
                                              args=(functions.lloyd_only_rref_p, et[iii], net[iii]),
                                              approx_grad=True, disp=False)

            refp  += [p]
            refii += [int((iii[0]+iii[-1])//2)]
    if len(refp) == 0:
        raise ValueError('Error nee2gpp_reichstein: No ref relationship found.')
        print('Warning nee2gpp_reichstein: No ref relationship found.')
        if masked:
            GPP  = np.ma.array(np.reshape(nee,inshape), mask=np.ones(inshape, dtype=bool))
            Reco = np.ma.array(np.reshape(nee,inshape), mask=np.ones(inshape, dtype=bool))
        else:
            GPP  = np.ones(np.reshape(nee,inshape))*undef
            Reco = np.ones(np.reshape(nee,inshape))*undef
        return GPP, Reco
    refp  = np.squeeze(np.array(refp))
    refii = np.squeeze(np.array(refii))

    # 4. Interpol Rref
    Rref = np.interp(dates, jul[refii], refp)

    # 5. Calc Reco
    Reco     = np.ones(ndata)*undef
    ii       = np.where(~t.mask)[0]
    Reco[ii] = functions.lloyd_fix(t[ii], Rref[ii], E0)

    # 6. Calc GPP
    GPP     = np.ones(ndata)*undef
    ii      = np.where(~(t.mask | nee.mask))[0]
    GPP[ii] = Reco[ii] - nee[ii]

    # 7. Set GPP=0 at night, if wanted
    if nogppnight:
        mask = isday | nee.mask | t.mask | isday.mask # night
        ii   = np.where(~mask)[0]
        Reco[ii] = nee[ii]
        GPP[ii]  = 0.
        # and prohibit negative gpp at any time
        mask = nee.mask | t.mask | (GPP>0.)
        ii   = np.where(~mask)[0]
        Reco[ii] -= GPP[ii]
        GPP[ii]  = 0.

    if masked:
        if np.isnan(undef):
            GPP  = np.ma.array(GPP,  mask=np.isnan(GPP))
            Reco = np.ma.array(Reco, mask=np.isnan(Reco))
        else:
            GPP  = np.ma.array(GPP,  mask=(GPP==undef))
            Reco = np.ma.array(Reco, mask=(Reco==undef))


    return GPP.reshape(inshape), Reco.reshape(inshape)

# ----------------------------------------------------------------------
def nee2gpp_lasslop(dates, nee, t, isday, rg, vpd, undef=np.nan,
                    shape=False, masked=False, nogppnight=False):
    """
        Calculate photosynthesis (GPP) and ecosystem respiration (Reco) from original
        Eddy flux data, using the daytime method of Lasslop et al. (2010),
        in order to calculate Reco and then GPP = Reco - NEE.


        Definition
        ----------
        def nee2gpp_lasslop(dates, nee, t, isday, rg, vpd, undef=np.nan,
                    shape=False, masked=False):


        Input
        -----
        Inputs are 1D arrays that can be masked or not.
        dates         julian days
        nee           net ecosystem exchange (uptake is <0) [umol m-2 s-1]
        t             temperature [K]
        rg            global radiation, i.e. shortwave down [W m-2]
        vpd           vapour pressure deficit [Pa]


        Parameters
        ----------
        undef        undefined values in data  (default: np.nan)
                     Input arrays will be masked at undef, keeping the original mask
        shape        if False then outputs are 1D arrays;
                     if True, output have the same shape as datain
                     if a shape tuple is given, then this tuple is used to reshape
        masked       if False: outputs are undef where nee and t are masked or undef
                     if True:  return masked arrays where outputs would be undef
        nogppnight   if True:  Resp=NEE, GPP=0 at night
                     if False: Resp=lloyd_taylor, GPP=Resp-NEE at night (default)


        Ouput
        -----
        GPP, Reco    photosynthesis, ecosystem respiration


        Restrictions
        ------------
        None.


        Literature
        ----------
        Lasslop et al. (2010)
            Separation of net ecosystem exchange into assimilation and respiration using
            a light response curve approach: critical issues and global evaluation
            Global Change Biology 16, 187-208


        Examples
        --------
        >>> from jams.fread import fread # from jams
        >>> from jams.date2dec import date2dec # from jams
        >>> dat   = fread('test_nee2gpp.csv', skip=2, transpose=True)
        >>> dates = date2dec(dy=dat[0,:], mo=dat[1,:], yr=dat[2,:], hr=dat[3,:], mi=dat[4,:])
        >>> NEE   = np.squeeze(dat[5,:])
        >>> rg    = np.squeeze(dat[6,:])
        >>> tair  = np.squeeze(dat[7,:])
        >>> undef = -9999.
        >>> isday = np.where(rg > 10., True, False)
        >>> tt    = np.where(tair == undef, undef, tair+273.15)
        >>> VPD = np.squeeze(dat[8,:])
        >>> vpd = np.where(VPD == undef, undef, VPD*100.)
        >>> # partition
        >>> GPP, Reco = nee2gpp(dates, NEE, tt, isday, rg, vpd, undef=undef, method='day')
        >>> print(GPP[1120:1128])
        [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  2.78457540e+00
          6.63212545e+00  8.88902165e+00  6.74243873e+00  9.51364527e+00]
        >>> print(Reco[1120:1128])
        [0.28786696 0.34594516 0.43893276 0.5495954  0.70029545 0.90849165
         1.15074873 1.46137527]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2013 Matthias Cuntz, Arndt Piayda - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written  MC, Mar 2012
        Modified AP, Mar 2012 - undef=np.nan
                 MC, Nov 2012 - individual routine
                 MC, Feb 2013 - ported to Python 3
    """

    # Checks

    # remember shape if any
    inshape = nee.shape
    dates   = np.squeeze(dates)
    nee     = np.squeeze(nee)
    t       = np.squeeze(t)
    isday   = np.squeeze(isday)
    # Check squeezed shape
    if dates.ndim != 1: raise ValueError('Error nee2gpp_lasslop: squeezed dates must be 1D array.')
    if nee.ndim   != 1: raise ValueError('Error nee2gpp_lasslop: squeezed nee must be 1D array.')
    if t.ndim     != 1: raise ValueError('Error nee2gpp_lasslop: squeezed t must be 1D array.')
    if isday.ndim != 1: raise ValueError('Error nee2gpp_lasslop: squeezed isday must be 1D array.')
    ndata = dates.size
    if ((nee.size != ndata) | (t.size != ndata) | (isday.size != ndata)):
        raise ValueError('Error nee2gpp_lasslop: inputs must have the same size.')
    if rg.ndim  != 1: raise ValueError('Error nee2gpp_lasslop: squeezed rg must be 1D array.')
    if vpd.ndim != 1: raise ValueError('Error nee2gpp_lasslop: squeezed vpd must be 1D array.')
    if ((rg.size != ndata) | (vpd.size != ndata)):
        raise ValueError('Error nee2gpp_lasslop: lasslop inputs must have the same size as other inputs.')

    # Transform to masked array with 1D mask
    nee   = np.ma.array(nee, mask=False)
    t     = np.ma.array(t, mask=False)
    isday = np.ma.array(isday, mask=False)
    rg    = np.ma.array(rg, mask=False)
    vpd   = np.ma.array(vpd, mask=False)
    # mask also undef
    if np.isnan(undef):
        if np.ma.any(np.isnan(nee)):   nee[np.isnan(nee)]     = np.ma.masked
        if np.ma.any(np.isnan(t)):     t[np.isnan(t)]         = np.ma.masked
        if np.ma.any(np.isnan(isday)): isday[np.isnan(isday)] = np.ma.masked
        if np.ma.any(np.isnan(rg)):    rg[np.isnan(rg)]       = np.ma.masked
        if np.ma.any(np.isnan(vpd)):   vpd[np.isnan(vpd)]     = np.ma.masked
    else:
    	if np.ma.any(nee==undef):   nee[nee==undef]     = np.ma.masked
    	if np.ma.any(t==undef):     t[t==undef]         = np.ma.masked
    	if np.ma.any(isday==undef): isday[isday==undef] = np.ma.masked
    	if np.ma.any(rg==undef):    rg[rg==undef]       = np.ma.masked
    	if np.ma.any(vpd==undef):   vpd[vpd==undef]     = np.ma.masked

    # Partition - Lasslop et al. (2010) method
    do_lgpp = False
    mask  = nee.mask | t.mask | isday.mask | rg.mask | vpd.mask
    # night
    nmask = isday | mask
    nii   = np.squeeze(np.where(~nmask))
    njul  = dates[nii]
    ntt   = np.ma.compressed(t[nii])
    nnet  = np.ma.compressed(nee[nii])
    aRref = np.mean(nnet)
    # day
    dmask = (~isday) | mask
    dii   = np.squeeze(np.where(~dmask))
    djul  = dates[dii]
    dtt   = np.ma.compressed(t[dii])
    dnet  = np.ma.compressed(nee[dii])
    drg   = np.ma.compressed(rg[dii])
    dvpd  = np.ma.compressed(vpd[dii])
    # starting values for optim
    aalpha = 0.01
    qnet   = np.sort(dnet)
    nqnet  = qnet.size
    abeta0 = np.abs(qnet[np.floor(0.97*nqnet).astype(int)]-qnet[np.ceil(0.03*nqnet).astype(int)])
    ak     = 0.
    # out
    lE0    = []
    lalpha = []
    if do_lgpp:
        lbeta0 = []
        lk     = []
    lRref  = []
    lii    = []
    dmin = np.floor(np.amin(dates)).astype(int)
    dmax = np.ceil(np.amax(dates)).astype(int)
    zaehl = -1
    for i in range(dmin,dmax,2):
        good = True
        # 1. Estimate E0 from nighttime data
        iii  = np.squeeze(np.where((njul>=i) & (njul<(i+12))))
        niii = iii.size
        if niii > 3:
            # p, c = opt.curve_fit(functions.lloyd_fix, ntt[iii], nnet[iii], p0=[aRref,100.])
            #p  = opt.fmin(functions.cost_lloyd_fix, [aRref,100.], args=(ntt[iii], nnet[iii]), disp=False)
            #p  = opt.fmin(functions.cost_abs, [aRref,100.], args=(functions.lloyd_fix_p, ntt[iii], nnet[iii]), disp=False)
            p, temp1, temp2 = opt.fmin_tnc(functions.cost_abs, [aRref,100.], bounds=[[0.,None],[0.,None]],
                                              args=(functions.lloyd_fix_p, ntt[iii], nnet[iii]),
                                              approx_grad=True, disp=False)

            E0 = np.maximum(p[1], 50.)
        else:
            if zaehl >= 0:
                E0 = lE0[zaehl]
            else:
                # large gap at beginning of data set, i.e. skip the period
                good = False
                continue
        # 2. Estimate alpha, k, beta0, Rref from daytime data
        iii  = np.squeeze(np.where((djul>=i) & (djul<(i+4))))
        niii = iii.size
        if niii > 3:
            et     = functions.lloyd_fix(dtt[iii], 1., E0)
            again  = True
            ialpha = aalpha
            ibeta0 = abeta0
            ik     = ak
            iRref  = aRref
            bounds = [[None,None],[None,None],[None,None],[None,None]]
            while again:
                again = False
                p, nfeval, rc  = opt.fmin_tnc(functions.cost_lasslop, [ialpha,ibeta0,ik,iRref], bounds=bounds,
                                              args=(drg[iii], et, dvpd[iii], dnet[iii]),
                                              approx_grad=True, disp=False)

                # if parameters beyond some bounds, set params and redo the optim or skip
                if ((p[0] < 0.) | (p[0] > 0.22)): # alpha
                    again = True
                    if zaehl >= 0:
                        bounds[0] = [lalpha[zaehl],lalpha[zaehl]]
                        ialpha    = lalpha[zaehl]
                    else:
                        bounds[0] = [0.,0.]
                        ialpha    = 0.
                if p[1] < 0.:                    # beta0
                    bounds[1] = [0.,0.]
                    ibeta0    = 0.
                    again = True
                if p[1] > 250.:
                    good = False
                    continue
                if p[2] < 0.:                    # k
                    bounds[2] = [0.,0.]
                    ik        = 0.
                    again = True
                if p[3] < 0:                     # Rref
                    good = False
                    continue
            if good:
                lalpha = lalpha + [p[0]]
                if do_lgpp:
                    lbeta0 = lbeta0 + [p[1]]
                    lk     = lk     + [p[2]]
                lRref  = lRref  + [p[3]]
                lii    = lii    + [int((iii[0]+iii[-1])/2)]
            else:
                continue
        else:
            continue
        lE0    = lE0 + [E0]
        zaehl += 1
    if len(lE0) == 0:
        raise ValueError('Error nee2gpp_lasslop: No day relationship found.')
    lE0    = np.squeeze(np.array(lE0))
    if do_lgpp:
        lalpha = np.squeeze(np.array(lalpha))
        lbeta0 = np.squeeze(np.array(lbeta0))
        lk     = np.squeeze(np.array(lk))
    lRref  = np.squeeze(np.array(lRref))
    lii    = np.squeeze(np.array(lii))

    # 3. Interpol E0 and Rref
    E0   = np.interp(dates, djul[lii], lE0)
    Rref = np.interp(dates, djul[lii], lRref)

    # 4. Calc Reco
    Reco     = np.ones(ndata)*undef
    ii       = np.squeeze(np.where(~t.mask))
    Reco[ii] = functions.lloyd_fix(t[ii], Rref[ii], E0[ii])

    # 5. Calc GPP from light response for check
    if do_lgpp:
        alpha    = np.interp(dates, djul[lii], lE0)
        beta0    = np.interp(dates, djul[lii], lbeta0)
        k        = np.interp(dates, djul[lii], lk)
        et       = functions.lloyd_fix(t, 1., E0)
        lmask    = t.mask | isday.mask | rg.mask | vpd.mask
        ii       = np.squeeze(np.where(~lmask))
        lgpp     = np.zeros(ndata)
        lgpp[ii] = functions.lasslop(rg[ii], et[ii], vpd[ii], alpha[ii], beta0[ii], k[ii], Rref[ii]) - Reco[ii]

    # 6. GPP
    GPP     = np.ones(ndata)*undef
    ii      = np.squeeze(np.where(~(t.mask | nee.mask)))
    GPP[ii] = Reco[ii] - nee[ii]

    # 7. Set GPP=0 at night, if wanted
    if nogppnight:
        mask = isday | nee.mask | t.mask | isday.mask # night
        ii   = np.where(~mask)[0]
        Reco[ii] = nee[ii]
        GPP[ii]  = 0.
        # and prohibit negative gpp at any time
        mask = nee.mask | t.mask | (GPP>0.)
        ii   = np.where(~mask)[0]
        Reco[ii] -= GPP[ii]
        GPP[ii]  = 0.


    if masked:
        if np.isnan(undef):
            GPP  = np.ma.array(GPP,  mask=np.isnan(GPP))
            Reco = np.ma.array(Reco, mask=np.isnan(Reco))
        else:
            GPP  = np.ma.array(GPP,  mask=(GPP == undef))
            Reco = np.ma.array(Reco, mask=(Reco == undef))

    if shape != False:
        if shape != True:
            return np.reshape(GPP,shape), np.reshape(Reco,shape)
        else:
            return np.reshape(GPP,inshape), np.reshape(Reco,inshape)
    else:
        return GPP, Reco


# -------------------------------------------------------------
if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # from jams.fread import fread # from jams
    # from jams.date2dec import date2dec # from jams
    # dat   = fread('test_nee2gpp.csv', skip=2, transpose=True)
    # dates = date2dec(dy=dat[0,:], mo=dat[1,:], yr=dat[2,:], hr=dat[3,:], mi=dat[4,:])
    # NEE   = np.squeeze(dat[5,:])
    # rg    = np.squeeze(dat[6,:])
    # tair  = np.squeeze(dat[7,:])
    # undef = -9999.
    # isday = np.where(rg > 10., True, False)
    # tt    = np.where(tair == undef, undef, tair+273.15)
    # GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='local')
    # print(GPP[1120:1128])
    # # [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.40606871e+00
    # #   8.31942152e+00  1.06242542e+01  8.49245664e+00  1.12381973e+01]
    # print(Reco[1120:1128])
    # # [1.68311981 1.81012431 1.9874173  2.17108871 2.38759152 2.64372415
    # #  2.90076664 3.18592735]
    # GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='local')
    # print(GPP[1120:1128])
    # # [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.40606871e+00
    # #   8.31942152e+00  1.06242542e+01  8.49245664e+00  1.12381973e+01]
    # GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='global')
    # print(GPP[1120:1128])
    # # [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  4.33166157e+00
    # #   8.18228013e+00  1.04092252e+01  8.19395317e+00  1.08427448e+01]
    # GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='Reichstein', masked=True)
    # print(GPP[1120:1128])
    # # [-- -- -- 4.406068706013192 8.319421516040766 10.624254150217764
    # #  8.492456637225963 11.238197347837367]
    # GPP, Reco = nee2gpp(dates, NEE, tt, isday, undef=undef, method='reichstein', shape=(np.size(NEE),1))
    # print(GPP[1120:1128])
    # # [[-9.99900000e+03]
    # #  [-9.99900000e+03]
    # #  [-9.99900000e+03]
    # #  [ 4.40606871e+00]
    # #  [ 8.31942152e+00]
    # #  [ 1.06242542e+01]
    # #  [ 8.49245664e+00]
    # #  [ 1.12381973e+01]]
    # VPD = np.squeeze(dat[8,:])
    # vpd = np.where(VPD == undef, undef, VPD*100.)
    # GPP, Reco = nee2gpp(dates, NEE, tt, isday, rg, vpd, undef=undef, method='day')
    # print(GPP[1120:1128])
    # # [-9.99900000e+03 -9.99900000e+03 -9.99900000e+03  2.78457540e+00
    # #   6.63212545e+00  8.88902165e+00  6.74243873e+00  9.51364527e+00]
    # print(Reco[1120:1128])
    # # [0.28786696 0.34594516 0.43893276 0.5495954  0.70029545 0.90849165
    # #  1.15074873 1.46137527]
